############################################################################
# tools/pynuttx/nxreg/register.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
from binascii import hexlify
from typing import List, Union

from construct import Array, Int8ul, Int16sl, Int16ul, PaddedString, Struct
from nxelf.elf import LiefELF

REGINFO_OFFSET_INVALID = -2
REGINFO_OFFSET_AUTO = -1

g_reg_table = {
    "arm": {
        "architecture": [
            "arm",
            "armv6-m",
            "armv6s-m",
            "armv7e-m",
            "armv8-m",
            "armv8-m.base",
            "armv8-m.main",
            "armv8.1-m.main",
        ],
        "feature": "org.gnu.gdb.arm.m-profile",
        "registers": [
            ("r0", 0, REGINFO_OFFSET_AUTO),
            ("r1", 1, REGINFO_OFFSET_AUTO),
            ("r2", 2, REGINFO_OFFSET_AUTO),
            ("r3", 3, REGINFO_OFFSET_AUTO),
            ("r4", 4, REGINFO_OFFSET_AUTO),
            ("r5", 5, REGINFO_OFFSET_AUTO),
            ("r6", 6, REGINFO_OFFSET_AUTO),
            ("r7", 7, REGINFO_OFFSET_AUTO),
            ("r8", 8, REGINFO_OFFSET_AUTO),
            ("r9", 9, REGINFO_OFFSET_AUTO),
            ("r10", 10, REGINFO_OFFSET_AUTO),
            ("r11", 11, REGINFO_OFFSET_AUTO),
            ("r12", 12, REGINFO_OFFSET_AUTO),
            ("sp", 13, REGINFO_OFFSET_AUTO),
            ("lr", 14, REGINFO_OFFSET_AUTO),
            ("pc", 15, REGINFO_OFFSET_AUTO),
            ("xpsr", 25, REGINFO_OFFSET_AUTO),
        ],
    },
    "arm-a": {
        "architecture": ["armv6", "armv7", "armv8-a", "armv8-r", "armv9-a"],
        "feature": "org.gnu.gdb.arm",
        "registers": [
            ("r0", 0, REGINFO_OFFSET_AUTO),
            ("r1", 1, REGINFO_OFFSET_AUTO),
            ("r2", 2, REGINFO_OFFSET_AUTO),
            ("r3", 3, REGINFO_OFFSET_AUTO),
            ("r4", 4, REGINFO_OFFSET_AUTO),
            ("r5", 5, REGINFO_OFFSET_AUTO),
            ("r6", 6, REGINFO_OFFSET_AUTO),
            ("r7", 7, REGINFO_OFFSET_AUTO),
            ("r8", 8, REGINFO_OFFSET_AUTO),
            ("r9", 9, REGINFO_OFFSET_AUTO),
            ("r10", 10, REGINFO_OFFSET_AUTO),
            ("r11", 11, REGINFO_OFFSET_AUTO),
            ("r12", 12, REGINFO_OFFSET_AUTO),
            ("sp", 13, REGINFO_OFFSET_AUTO),
            ("lr", 14, REGINFO_OFFSET_AUTO),
            ("pc", 15, REGINFO_OFFSET_AUTO),
            ("cpsr", 25, 164),
        ],
    },
    "arm64": {
        "architecture": ["aarch64", "aarch64:ilp32", "aarch64:armv8-r"],
        "feature": "org.gnu.gdb.aarch64",
        "registers": [
            ("x0", 0, REGINFO_OFFSET_AUTO),
            ("x1", 1, REGINFO_OFFSET_AUTO),
            ("x2", 2, REGINFO_OFFSET_AUTO),
            ("x3", 3, REGINFO_OFFSET_AUTO),
            ("x4", 4, REGINFO_OFFSET_AUTO),
            ("x5", 5, REGINFO_OFFSET_AUTO),
            ("x6", 6, REGINFO_OFFSET_AUTO),
            ("x7", 7, REGINFO_OFFSET_AUTO),
            ("x8", 8, REGINFO_OFFSET_AUTO),
            ("x9", 9, REGINFO_OFFSET_AUTO),
            ("x10", 10, REGINFO_OFFSET_AUTO),
            ("x11", 11, REGINFO_OFFSET_AUTO),
            ("x12", 12, REGINFO_OFFSET_AUTO),
            ("x13", 13, REGINFO_OFFSET_AUTO),
            ("x14", 14, REGINFO_OFFSET_AUTO),
            ("x15", 15, REGINFO_OFFSET_AUTO),
            ("x16", 16, REGINFO_OFFSET_AUTO),
            ("x17", 17, REGINFO_OFFSET_AUTO),
            ("x18", 18, REGINFO_OFFSET_AUTO),
            ("x19", 19, REGINFO_OFFSET_AUTO),
            ("x20", 20, REGINFO_OFFSET_AUTO),
            ("x21", 21, REGINFO_OFFSET_AUTO),
            ("x22", 22, REGINFO_OFFSET_AUTO),
            ("x23", 23, REGINFO_OFFSET_AUTO),
            ("x24", 24, REGINFO_OFFSET_AUTO),
            ("x25", 25, REGINFO_OFFSET_AUTO),
            ("x26", 26, REGINFO_OFFSET_AUTO),
            ("x27", 27, REGINFO_OFFSET_AUTO),
            ("x28", 28, REGINFO_OFFSET_AUTO),
            ("x29", 29, REGINFO_OFFSET_AUTO),
            ("x30", 30, REGINFO_OFFSET_AUTO),
            ("sp", 31, REGINFO_OFFSET_AUTO),  # SP
            ("pc", 32, REGINFO_OFFSET_AUTO),  # PC
        ],
    },
    "riscv": {
        "architecture": ["riscv", "riscv:rv32", "riscv:rv64"],
        "feature": "org.gnu.gdb.riscv:rv32",
        "registers": [
            ("zero", 0, REGINFO_OFFSET_AUTO),
            ("ra", 1, REGINFO_OFFSET_AUTO),
            ("sp", 2, REGINFO_OFFSET_AUTO),
            ("gp", 3, REGINFO_OFFSET_AUTO),
            ("tp", 4, REGINFO_OFFSET_AUTO),
            ("t0", 5, REGINFO_OFFSET_AUTO),
            ("t1", 6, REGINFO_OFFSET_AUTO),
            ("t2", 7, REGINFO_OFFSET_AUTO),
            ("fp", 8, REGINFO_OFFSET_AUTO),
            ("s1", 9, REGINFO_OFFSET_AUTO),
            ("a0", 10, REGINFO_OFFSET_AUTO),
            ("a1", 11, REGINFO_OFFSET_AUTO),
            ("a2", 12, REGINFO_OFFSET_AUTO),
            ("a3", 13, REGINFO_OFFSET_AUTO),
            ("a4", 14, REGINFO_OFFSET_AUTO),
            ("a5", 15, REGINFO_OFFSET_AUTO),
            ("a6", 16, REGINFO_OFFSET_AUTO),
            ("a7", 17, REGINFO_OFFSET_AUTO),
            ("s2", 18, REGINFO_OFFSET_AUTO),
            ("s3", 19, REGINFO_OFFSET_AUTO),
            ("s4", 20, REGINFO_OFFSET_AUTO),
            ("s5", 21, REGINFO_OFFSET_AUTO),
            ("s6", 22, REGINFO_OFFSET_AUTO),
            ("s7", 23, REGINFO_OFFSET_AUTO),
            ("s8", 24, REGINFO_OFFSET_AUTO),
            ("s9", 25, REGINFO_OFFSET_AUTO),
            ("s10", 26, REGINFO_OFFSET_AUTO),
            ("s11", 27, REGINFO_OFFSET_AUTO),
            ("t3", 28, REGINFO_OFFSET_AUTO),
            ("t4", 29, REGINFO_OFFSET_AUTO),
            ("t5", 30, REGINFO_OFFSET_AUTO),
            ("t6", 31, REGINFO_OFFSET_AUTO),
            ("pc", 33, REGINFO_OFFSET_AUTO),  # PC
        ],
    },
    "x86": {
        "architecture": ["i386", "i386:intel"],
        "feature": "org.gnu.gdb.i386:x86",
        "registers": [
            ("eax", 0, REGINFO_OFFSET_AUTO),
            ("ecx", 1, REGINFO_OFFSET_AUTO),
            ("edx", 2, REGINFO_OFFSET_AUTO),
            ("ebx", 3, REGINFO_OFFSET_AUTO),
            ("esp", 4, REGINFO_OFFSET_AUTO),
            ("ebp", 5, REGINFO_OFFSET_AUTO),
            ("esi", 6, REGINFO_OFFSET_AUTO),
            ("edi", 7, REGINFO_OFFSET_AUTO),
            ("eip", 8, REGINFO_OFFSET_AUTO),
            ("eflags", 9, REGINFO_OFFSET_AUTO),
            ("cs", 10, REGINFO_OFFSET_AUTO),
            ("ss", 11, REGINFO_OFFSET_AUTO),
            ("ds", 12, REGINFO_OFFSET_AUTO),
            ("es", 13, REGINFO_OFFSET_AUTO),
            ("fs", 14, REGINFO_OFFSET_AUTO),
            ("gs", 15, REGINFO_OFFSET_AUTO),
        ],
    },
    "x86-64": {
        "architecture": ["i386:x86-64", "i386:x86-64:intel"],
        "feature": "org.gnu.gdb.i386:x86-64",
        "registers": [
            ("rax", 0, REGINFO_OFFSET_AUTO),
            ("rbx", 1, REGINFO_OFFSET_AUTO),
            ("rcx", 2, REGINFO_OFFSET_AUTO),
            ("rdx", 3, REGINFO_OFFSET_AUTO),
            ("rsi", 4, REGINFO_OFFSET_AUTO),
            ("rdi", 5, REGINFO_OFFSET_AUTO),
            ("rbp", 6, REGINFO_OFFSET_AUTO),
            ("rsp", 7, REGINFO_OFFSET_AUTO),
            ("r8", 8, REGINFO_OFFSET_AUTO),
            ("r9", 9, REGINFO_OFFSET_AUTO),
            ("r10", 10, REGINFO_OFFSET_AUTO),
            ("r11", 11, REGINFO_OFFSET_AUTO),
            ("r12", 12, REGINFO_OFFSET_AUTO),
            ("r13", 13, REGINFO_OFFSET_AUTO),
            ("r14", 14, REGINFO_OFFSET_AUTO),
            ("r15", 15, REGINFO_OFFSET_AUTO),
            ("rip", 16, REGINFO_OFFSET_AUTO),
            ("eflags", 17, REGINFO_OFFSET_AUTO),
            ("cs", 18, REGINFO_OFFSET_AUTO),
            ("ss", 19, REGINFO_OFFSET_AUTO),
            ("ds", 20, REGINFO_OFFSET_AUTO),
            ("es", 21, REGINFO_OFFSET_AUTO),
            ("fs", 22, REGINFO_OFFSET_AUTO),
        ],
    },
    "esp32s3": {
        "architecture": ["esp32s3"],  # Use xtensa-esp32s3-elf-gdb
        "feature": "",
        "registers": [
            ("pc", 0, 0),
            ("ps", 73, 292, 0x40000),  # g_reg_offs placed it in the second position
            ("a0", 1, 4),
            ("a1", 2, REGINFO_OFFSET_AUTO),
            ("a2", 3, REGINFO_OFFSET_AUTO),
            ("a3", 4, REGINFO_OFFSET_AUTO),
            ("a4", 5, REGINFO_OFFSET_AUTO),
            ("a5", 6, REGINFO_OFFSET_AUTO),
            ("a6", 7, REGINFO_OFFSET_AUTO),
            ("a7", 8, REGINFO_OFFSET_AUTO),
            ("a8", 9, REGINFO_OFFSET_AUTO),
            ("a9", 10, REGINFO_OFFSET_AUTO),
            ("a10", 11, REGINFO_OFFSET_AUTO),
            ("a11", 12, REGINFO_OFFSET_AUTO),
            ("a12", 13, REGINFO_OFFSET_AUTO),
            ("a13", 14, REGINFO_OFFSET_AUTO),
            ("a14", 15, REGINFO_OFFSET_AUTO),
            ("a15", 16, REGINFO_OFFSET_AUTO),
            ("windowbase", 69, 276, 0),
            ("windowstart", 70, 280, 1),
        ],
    },
    "xtensa": {
        "architecture": ["xtensa"],  # Use xt-gdb
        "feature": "",
        "registers": [
            ("pc", 32, 0),
            ("ps", 742, 472, 0x40000),
            ("a0", 256, 4),
            ("a1", 257, REGINFO_OFFSET_AUTO),
            ("a2", 258, REGINFO_OFFSET_AUTO),
            ("a3", 259, REGINFO_OFFSET_AUTO),
            ("a4", 260, REGINFO_OFFSET_AUTO),
            ("a5", 261, REGINFO_OFFSET_AUTO),
            ("a6", 262, REGINFO_OFFSET_AUTO),
            ("a7", 263, REGINFO_OFFSET_AUTO),
            ("a8", 264, REGINFO_OFFSET_AUTO),
            ("a9", 265, REGINFO_OFFSET_AUTO),
            ("a10", 266, REGINFO_OFFSET_AUTO),
            ("a11", 267, REGINFO_OFFSET_AUTO),
            ("a12", 268, REGINFO_OFFSET_AUTO),
            ("a13", 269, REGINFO_OFFSET_AUTO),
            ("a14", 270, REGINFO_OFFSET_AUTO),
            ("a15", 271, REGINFO_OFFSET_AUTO),
            ("windowbase", 584, 308, 0),
            ("windowstart", 585, 312, 1),
        ],
    },
    "tricore": {
        "architecture": [
            "tricore:v1.2",
            "tricore:v1.1",
            "tricore:v1.3",
            "tricore:v1.3.1",
            "tricore:v1_6",
            "tricore:v1_6_1",
            "tricore:v1_6_2",
            "tricore:v1_8",
        ],  # Use tricore-gdb
        "feature": "",
        "registers": [
            ("pcx", 34, 136),
            ("pc", 36, 144),
            ("d0", 0, 0),
            ("d1", 1, 4),
            ("d2", 2, 8),
            ("d3", 3, 12),
            ("d4", 4, 16),
            ("d5", 5, 20),
            ("d6", 6, 24),
            ("d7", 7, 28),
            ("a2", 18, 72),
            ("a3", 19, 76),
            ("a4", 20, 80),
            ("a5", 21, 84),
            ("a6", 22, 88),
            ("a7", 23, 92),
            ("psw", 35, 140),
            ("sp", 26, 104),
            ("a11", 27, 108),
            ("d8", 8, 32),
            ("d9", 9, 36),
            ("d10", 10, 40),
            ("d11", 11, 44),
            ("a12", 28, 112),
            ("a13", 29, 116),
            ("a14", 30, 120),
            ("a15", 31, 124),
            ("d12", 12, 48),
            ("d13", 13, 52),
            ("d14", 14, 56),
            ("d15", 15, 60),
        ],
    },
}


class RegInfo:
    def __init__(self, name, size, toffset=0):
        self.name = name
        self.size = size
        self.toffset = toffset

    def __str__(self):
        return f"{self.name}({self.size})"

    def __repr__(self):
        return f"REG({self.name}, {self.size}, {self.toffset})"


def get_reginfo(elf: LiefELF) -> List[RegInfo]:
    # Now get register offset in TCB
    _, data = elf.read_symbol("g_reginfo")

    try:
        reginfo_s = Struct(
            "name" / PaddedString(8, "utf-8"),
            "size" / Int8ul,
            "regnum" / Int8ul,
            "toffset" / Int16sl,
            "goffset" / Int16sl,
            "reserved" / Int16ul,
        )
        regsnum = len(data) // reginfo_s.sizeof()
        reginfo_t = Array(regsnum, reginfo_s)
        reginfo = reginfo_t.parse(data)
    except Exception:

        reginfo_s = Struct(
            "name" / PaddedString(8, "utf-8"),
            "size" / Int8ul,
            "regnum" / Int8ul,
            "toffset" / Int16sl,
            "goffset" / Int16sl,
        )
        regsnum = len(data) // reginfo_s.sizeof()
        reginfo_t = Array(regsnum, reginfo_s)
        reginfo = reginfo_t.parse(data)

    return [RegInfo(reg.name, elf.bits // 8, reg.toffset) for reg in reginfo]


class Register:
    def __init__(
        self, name, regnum, size: int, goffset=0, toffset=0, value=0, fixedvalue=None
    ):
        self.name = name
        self.regnum = regnum
        self.size = size  # size in bytes
        assert (
            goffset != REGINFO_OFFSET_AUTO
        )  # Must be the final processed offset for GDB g/G packet
        self.goffset = goffset
        self.toffset = toffset
        self._value = value
        self.fixedvalue = fixedvalue
        self.logger = logging.getLogger(__name__)

    def __str__(self):
        return f"{self.name}({self.regnum}, size:{self.size}, offset:{self.goffset},{self.toffset} value:{self.value:#x})"

    def __repr__(self):
        return self.__str__()

    def __bytes__(self):
        return self.value.to_bytes(self.size, "little", signed=self._value < 0)

    @property
    def has_value(self):
        return self._value is not None or self.fixedvalue is not None

    @property
    def value(self):
        return self._value if self.fixedvalue is None else self.fixedvalue

    @value.setter
    def value(self, value):
        """Set register value from bytes or value"""
        if self.name == "":
            return

        if isinstance(value, bytes) or isinstance(value, bytearray):
            if len(value) != self.size:
                raise ValueError(
                    f"Invalid value, expected {self.size} bytes, got {len(value)} bytes"
                )
            self._value = int.from_bytes(bytes(value), "little")
        elif isinstance(value, int):
            self._value = value
        else:
            raise ValueError(f"Invalid value type: {type(value)}")
        self.logger.debug(f"Set {self.name} = {self._value:#x}")


class Registers:
    def __init__(self, elf: LiefELF, arch=None, readmem=None):
        """
        Registers class to store register information

        :param elf: Parsed ELF file
        :param arch: architecture name, or use current gdb architecture by default
        :param readmem: function to read memory
        """
        # if we don't have register names in elf, fallback to hardcoded register layouts
        if not arch:
            raise ValueError("Architecture is required to get register names")

        self.logger = logging.getLogger(__name__)
        self.arch = arch
        self.readmem = readmem
        self._registers: List[Register] = []
        regsize = elf.get_pointer_size()

        reginfo = get_reginfo(elf)
        layouts = g_reg_table.get(self.arch, {}).get("registers", [])
        goffset = 0
        for i, (name, regnum, offset, *fixed) in enumerate(layouts):
            offset = goffset if offset == REGINFO_OFFSET_AUTO else offset
            goffset = offset + regsize  # move to next register offset

            info = next((r for r in reginfo if r.name == name), None)
            has_toffset = info and info.toffset != REGINFO_OFFSET_INVALID
            if not (fixed or has_toffset):
                # Not a fixed fake register, nor a register in TCB
                self.logger.debug(f"Register {name}({regnum}) not found in TCB")
                continue

            register = Register(
                name=name,
                regnum=regnum,
                size=regsize,
                goffset=offset,
                toffset=info.toffset if has_toffset else 0,
                value=None if not has_toffset else 0,
                fixedvalue=fixed[0] if fixed else None,
            )

            self._registers.append(register)
            self.logger.debug(
                f"Register {name}({regnum}) offset: {offset}, tcb_off: {register.toffset}"
            )

        self._registers.sort(key=lambda x: x.regnum)
        ncpu = elf.get_symbol("g_running_tasks").size // elf.get_pointer_size()
        self.xcp_reg_size = elf.get_symbol("g_last_regs").size // ncpu

    def __str__(self):
        return f"({self.arch} x{len(self._registers)}, {self.sizeof()} bytes)"

    def __repr__(self):
        return f"({self.arch} x{len(self._registers)}, {self.sizeof()} bytes)"

    def sizeof(self):
        """Return total register size in byte"""
        return sum(r.size for r in self._registers)

    def get(self, regnum: int = None, name: str = None) -> Register:
        """Get register by register number"""
        for reg in self._registers:
            if reg.regnum == regnum:
                return reg

            if reg.name == name:
                return reg

        return None

    def set(self, value: Union[int, bytes], regnum: int = None, name: str = None):
        """Set register value by register number"""
        reg = self.get(regnum=regnum, name=name)
        if not reg:
            raise KeyError(f"Register {name or regnum} not found")
        reg.value = value

    def load(self, addr: int):
        """
        Load register values from various sources.

        :param tcb: load register values from tcb.xcp.regs
        :param address: load register values from specified address which points to regs in context
        """

        if self.readmem is None:
            raise ValueError("readmem is not set")

        xcpregs = self.readmem(addr, self.xcp_reg_size)
        if xcpregs:
            for reg in self._registers:
                reg.value = xcpregs[reg.toffset : reg.toffset + reg.size]
        else:
            raise ValueError("No valid source to load register values.\n")
        return self  # allow to build and use Register().load() directly

    def to_g(self):
        """Return GDB RSP g packet"""
        reply = b""
        goffset = 0
        for reg in self._registers:
            self.logger.debug(
                f"Register {reg.name}({reg.regnum}) at {reg.goffset}, "
                f"size: {reg.size}, value: {hex(reg.value)}, "
                f"fixed: {reg.fixedvalue}, has_value: {reg.has_value}"
            )

            if reg.goffset != goffset:
                reply += b"xx" * (reg.goffset - goffset)

            if not reg.has_value:
                reply += b"xx" * reg.size
            else:
                reply += hexlify(bytes(reg))

            goffset = reg.goffset + reg.size
        return reply

    def from_g(self, data: bytes):
        """Parse GDB RSP G packet"""
        for reg in self._registers:
            goffset = reg.goffset
            self.logger.debug(
                f"Parse {reg.name}({reg.regnum}) from {goffset}, data: {data[goffset:goffset+reg.size]}"
            )
            reg.value = data[goffset : goffset + reg.size]
            goffset = reg.goffset + reg.size

    def __iter__(self):
        return iter(self._registers)

    def __len__(self):
        return len(self._registers)

    def __getitem__(self, key):
        return self._registers[key]


def get_arch_name():
    """
    GDB uses arch name like aarch64, while we use arm64 to identify the register set.
    This function maps the GDB arch name to our arch name.
    """
    import gdb

    gdb_arch = gdb.selected_inferior().architecture()
    gdb_arch = gdb_arch.name().lower()

    for arch_key, arch_info in g_reg_table.items():
        if gdb_arch in arch_info["architecture"]:
            if arch_key == "xtensa" and "esp-gdb" in gdb.execute(
                "show version", to_string=True
            ):
                return "esp32s3"
            else:
                return arch_key
    return None



import argparse
import re
import shlex
from typing import List, Union

import gdb

from . import autocompeletion, utils

CONFIG_LIBC_BACKTRACE_DEPTH = utils.get_field_nitems(
    "struct backtrace_entry_s", "stack"
)


class Backtrace:
    """
    Convert addresses to backtrace
    Usage:
    backtrace = Backtrace(addresses=[0x4001, 0x4002, 0x4003])

    # Access converted backtrace
    addr, func, file, line = backtrace[0]
    remaining = backtrace[1:]  # Return list of (addr, func, file, line)

    # Iterate over backtrace
    for addr, func, file, line in backtrace:
        print(addr, func, file, line)

    # Append more addresses to convert
    backtrace.append(0x40001234)

    # Print backtrace
    print(str(backtrace))

    # Format backtrace to string
    print("\n".join(backtrace.formatted))

    # Custom formatter
    backtrace = Backtrace(addresses=[0x4001, 0x4002, 0x4003], formatter="{:<6} {:<20} {}")
    """

    def __init__(
        self,
        address: List[Union[gdb.Value, int]] = [],
        formatter="{:<5} {:<36} {}\n",
        break_null=True,
    ):
        self.formatter = formatter  # Address, Function, Source
        self._formatted = None  # Cached formatted backtrace
        self.backtrace = []
        for addr in address:
            if break_null and not addr:
                break
            self.append(int(addr))

    def __eq__(self, value: "Backtrace") -> bool:
        return self.backtrace == value.backtrace

    def __hash__(self) -> int:
        return hash(tuple(self.backtrace))

    def append(self, addr: Union[gdb.Value, int]) -> None:
        """Append an address to the backtrace"""
        if result := utils.Symbol(addr):
            self.backtrace.append(
                (result.address, result.func, result.filename, result.line)
            )
            self._formatted = None  # Clear cached result

    @property
    def formatted(self):
        """Return the formatted backtrace string list"""
        if not self._formatted:
            self._formatted = [
                self.formatter.format(hex(addr), func, f"{file}:{line}")
                for addr, func, file, line in self.backtrace
            ]

        return self._formatted

    def __repr__(self) -> str:
        return f"Backtrace: {len(self.backtrace)} items"

    def __str__(self) -> str:
        return "".join(self.formatted)

    def __iter__(self):
        for item in self.backtrace:
            yield item

    def __getitem__(self, index):
        return self.backtrace.__getitem__(index)

    def toJSON(self):
        return [
            {"address": addr, "function": func, "source": source, "line": line}
            for addr, func, source, line in self.backtrace
        ]


class BacktraceEntry:
    backtrace_entry_s = utils.lookup_type("struct backtrace_entry_s")

    def __init__(self, entry):
        if int(entry) == 0:
            self.entry = None
        else:
            self.entry = utils.Value(entry).cast(self.backtrace_entry_s.pointer())

    def get(self):
        if not self.entry:
            return list()

        stack = utils.ArrayIterator(self.entry.stack, self.entry.depth)
        return [int(addr) for addr in stack]

    def format(self, formatter: str = "{:<5} {:<36} {}\n"):
        if not self.entry:
            return ""

        stack = utils.ArrayIterator(self.entry.stack, self.entry.depth)
        return str(Backtrace(stack))


@autocompeletion.complete
class Addr2Line(gdb.Command):
    """Convert addresses or expressions

    Usage: addr2line address1 address2 expression1
    Example: addr2line 0x1234 0x5678
             addr2line "0x1234 + pointer->abc" &var var->field function_name var
             addr2line $pc $r1 "$r2 + var"
             addr2line [24/08/29 20:51:02] [CPU1] [209] [ap] sched_dumpstack: backtrace| 0: 0x402cd484 0x4028357e
             addr2line -f crash.log
             addr2line -f crash.log -p 123
    """

    formatter = "{:<20} {:<32} {}\n"

    def get_argparser(self):
        parser = argparse.ArgumentParser(
            description="Convert addresses or expressions to source code location"
        )
        parser.add_argument(
            "-f", "--file", type=str, metavar="file", help="Crash log to analyze."
        )
        parser.add_argument(
            "-p",
            "--pid",
            type=int,
            help="Only dump specified task backtrace from crash file.",
        )
        return parser

    def __init__(self):
        super().__init__("addr2line", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def print_backtrace(self, addresses, pid=None):
        if pid:
            gdb.write(f"\nBacktrace of {pid}\n")
        backtraces = Backtrace(addresses, formatter=self.formatter, break_null=False)
        gdb.write(str(backtraces))

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        if not args:
            gdb.write(Addr2Line.__doc__ + "\n")
            return
        pargs = None
        try:
            pargs, _ = self.parser.parse_known_args(gdb.string_to_argv(args))
        except SystemExit:
            pass

        gdb.write(self.formatter.format("Address", "Symbol", "Source"))

        if pargs and pargs.file:
            pattern = re.compile(
                r".*sched_dumpstack: backtrace\|\s*(\d+)\s*:\s*((?:(0x)?[0-9a-fA-F]+\s*)+)"
            )
            addresses = {}
            with open(pargs.file, "r") as f:
                for line in f:
                    match = pattern.match(line)
                    if not match:
                        continue

                    pid = match.group(1)
                    if pargs.pid is not None and pargs.pid != int(pid):
                        continue

                    addresses.setdefault(pid, [])
                    addresses[pid].extend(
                        [int(addr, 16) for addr in match.group(2).split()]
                    )

            for pid, addr in addresses.items():
                self.print_backtrace(addr, pid)
        else:
            addresses = []
            for arg in shlex.split(args.replace(",", " ")):
                if utils.is_decimal(arg):
                    addresses.append(int(arg))
                elif utils.is_hexadecimal(arg):
                    addresses.append(int(arg, 16))
                else:
                    try:
                        var = utils.parse_and_eval(f"{arg}")
                        addresses.append(var)
                    except gdb.error as e:
                        gdb.write(f"Ignore {arg}: {e}\n")
            self.print_backtrace(addresses)


@autocompeletion.complete
class BacktracePool(gdb.Command):
    """Display the global backtrace information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument("-d", "--detail", action="store_true")
        parser.add_argument("-t", "--top", type=int, help="Display the top N backtrace")
        return parser

    def __init__(self):
        super().__init__("backtracepool", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def invoke(self, args, from_tty):
        g_backtrace_pool = utils.parse_and_eval("g_backtrace_pool")
        args = self.parser.parse_args(gdb.string_to_argv(args))

        formatter = "{:>8} {:>8} {:>10} {:}\n"
        btformat = formatter.format("", "", "", "")[:-1] + "{1:<48}{2}\n"
        gdb.write(formatter.format("slot", "depth", "refcount", "stack"))

        backtrace_pool = []
        for i in range(g_backtrace_pool.capacity):
            if not g_backtrace_pool.bucket[i]:
                continue
            entry = BacktraceEntry(g_backtrace_pool.bucket[i])
            entry.index = i
            backtrace_pool.append(entry)

        backtrace_pool.sort(key=lambda x: x.entry.ref, reverse=True)

        if args.top:
            backtrace_pool = backtrace_pool[: args.top]

        for entry in backtrace_pool:
            stack = ""
            if args.detail:
                stack = f"\n{entry.format(btformat)}"
            else:
                stack = " ".join(hex(addr) for addr in entry.get())
            gdb.write(
                formatter.format(entry.index, entry.entry.depth, entry.entry.ref, stack)
            )

        gdb.write(
            f"capacity: {g_backtrace_pool.capacity}, used: {g_backtrace_pool.used}\n"
        )



############################################################################
# tools/pynuttx/tests/test_runtime_memory.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import re
import unittest

import gdb

# The following test cases require running the program as
# we need to access the memory of the program


class TestMemory(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def check_output(self, out, expect=r"Total \d+ blks, \d+ bytes"):
        if not re.compile(expect).search(out):
            self.fail(f"Got: \n{out}")

    def test_memdump(self):
        out = gdb.execute("memdump", to_string=True)
        self.check_output(out)

    def test_memdump_pid(self):
        out = gdb.execute("memdump -p 1", to_string=True)
        self.check_output(out)

    def test_memdump_addr(self):
        heap_start = gdb.parse_and_eval("g_mmheap")["mm_heapstart"][0]
        out = gdb.execute(f"memdump -a {hex(heap_start)}", to_string=True)
        self.check_output(out, expect=r"found belongs to")

    def test_memdump_free(self):
        out = gdb.execute("memdump --free", to_string=True)
        self.check_output(out)

    def test_memleak(self):
        out = gdb.execute("memleak", to_string=True)
        self.check_output(out, expect=r"Leaked \d+ blks, \d+ bytes")

    # memmap may stuck because of huge 2GB memory qemu provides.
    # def test_memmap(self):
    #     out = gdb.execute("memmap", to_string=True)

    def test_memfrag(self):
        out = gdb.execute("memfrag", to_string=True)
        self.check_output(out, expect=r"fragmentation rate")

    def test_mempool(self):
        out = gdb.execute("mempool", to_string=True)
        self.check_output(out, expect=r"Total \d+ pools")

    def test_mm_heap(self):
        out = gdb.execute("mm heap", to_string=True)
        # Umem@0x21023230, 2101106.4kB 2regions: 0x21023638~0x213fffd0,0x60000000~0xdfffffd0
        #  - has 21 nodes, regions:0x21023638~0x213fffd0 0x60000000~0xdfffffd0
        self.check_output(out, expect=r"Umem@0x[0-9a-f]+, \d+\.\d+kB \d+regions")



############################################################################
# tools/pynuttx/nxgdb/pm.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from . import utils

PM_COUNT = utils.get_field_nitems("struct pm_domain_s", "wakelock")


class PMStateInfo:
    def __init__(self, wake_time, sleep_time):
        self.wake_time = wake_time
        self.sleep_time = sleep_time
        self.total_time = wake_time + sleep_time


def has_inferior():
    inferior = gdb.selected_inferior()
    return inferior is not None and inferior.is_valid()


def get_pm_state_info_list(domain):
    state_info_list = []
    sum_time = 0

    for state in range(PM_COUNT):
        wake_time = int(domain["wake"][state]["tv_sec"])
        sleep_time = int(domain["sleep"][state]["tv_sec"])

        sum_time += wake_time + sleep_time
        state_info_list.append(PMStateInfo(wake_time, sleep_time))

    sum_time = sum_time if sum_time > 0 else 1
    return state_info_list, sum_time


class Pmconfig(gdb.Command):
    """Display power management configuration information

    Usage: pmconfig
    """

    state_formatter = "{:<12} {:<12} {:<12} {:<12}"
    state_header = ("WAKE", "SLEEP", "TOTAL")

    def __init__(self):
        if PM_COUNT:
            super().__init__("pmconfig", gdb.COMMAND_USER)

    def get_time_str(self, time_val, sum_time):
        percentage = (time_val / sum_time) * 100 if sum_time > 0 else 0
        return f"{int(time_val)}s {int(percentage):02}%"

    def print_state_info(self, domain_name, domain):
        state_path = f"/proc/pm/state{domain_name[-1]}"
        gdb.write(f"{state_path}:\n")
        gdb.write(self.state_formatter.format(domain_name, *self.state_header) + "\n")

        state_info_list, sum_time = get_pm_state_info_list(domain)
        pm_state = utils.parse_and_eval("g_pm_state")

        for state in range(PM_COUNT):
            state_info = state_info_list[state]
            wake_str = self.get_time_str(state_info.wake_time, sum_time)
            sleep_str = self.get_time_str(state_info.sleep_time, sum_time)
            total_str = self.get_time_str(state_info.total_time, sum_time)

            state_name = pm_state[state].string()
            gdb.write(
                self.state_formatter.format(state_name, wake_str, sleep_str, total_str)
                + "\n"
            )

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        if not has_inferior():
            return

        pmdomains = utils.parse_and_eval("g_pmdomains")
        ndomains = utils.nitems(pmdomains)

        domains = [f"DOMAIN{i}" for i in range(ndomains)]
        for i, domain_name in enumerate(domains):
            domain = pmdomains[i]
            self.print_state_info(domain_name, domain)

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Power Manager Information",
            "summary": "Power manager information",
            "command": "pmconfig",
            "result": "info",
            "category": utils.DiagnoseCategory.power,
            "message": gdb.execute("pmconfig", to_string=True),
        }



#! /usr/bin/python
############################################################################
# tools/pynuttx/tests/runner.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import os
import unittest

if __name__ == "__main__":
    test_dir = os.path.dirname(os.path.abspath(__file__))
    testsuit_runtime = unittest.defaultTestLoader.discover(
        test_dir, pattern="test_runtime*.py"
    )

    test_runner = unittest.TextTestRunner()

    # NOTE: Run the runtime testsuit first, as we are going to mock
    # lots of methods as well as classes later!
    result_runtime = test_runner.run(testsuit_runtime)

    """ testsuit_mock = unittest.defaultTestLoader.discover(
        test_dir, pattern="test_mock*.py"
    )
    result_mock = test_runner.run(testsuit_mock) """

    if not result_runtime.wasSuccessful():
        print("GDB Plugin Test Failed")
    else:
        print("GDB Plugin Test Successfully")



############################################################################
# tools/pynuttx/nxgdb/fs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb
import nxgdb.backtrace as bt
import nxgdb.protocols.fs as p
from nxgdb.fs.inode import inode_getpath
from nxgdb.fs.utils import CONFIG_FS_BACKTRACE, CONFIG_NFILE_DESCRIPTORS_PER_BLOCK
from nxgdb.protocols.thread import Tcb


def get_file(tcb: Tcb, fd):
    group = tcb.group
    fdlist = group.tg_fdlist
    fl_fds = fdlist.fl_fds
    fl_rows = fdlist.fl_rows

    row = fd // CONFIG_NFILE_DESCRIPTORS_PER_BLOCK
    col = fd % CONFIG_NFILE_DESCRIPTORS_PER_BLOCK

    if row >= fl_rows:
        return None

    return fl_fds[row][col]


def print_file_info(fd, fdp: p.Fd, file: p.File, formatter: str):
    backtrace_formatter = "{0:<5} {1:<36} {2}"

    oflags = int(file.f_oflags)
    pos = int(file.f_pos)
    path = inode_getpath(file.f_inode)

    output = []
    if CONFIG_FS_BACKTRACE:
        backtrace = tuple(bt.BacktraceEntry(fdp.f_backtrace).get())
        backtrace = bt.Backtrace(backtrace, backtrace_formatter)

        output.append(
            formatter.format(fd, hex(file), oflags, pos, path, backtrace.formatted[0])
        )
        output.extend(
            formatter.format("", "", "", "", "", bt) for bt in backtrace.formatted[1:]
        )
        output.append("")  # separate each backtrace
    else:
        output = [formatter.format(fd, hex(file), hex(oflags), pos, path, "")]

    gdb.write("\n".join(output).rstrip())  # strip trailing spaces
    gdb.write("\n")


def print_fdinfo_by_tcb(tcb, processed_groups, total_fd_count):
    """print fdlist from tcb"""
    gdb.write(f"PID: {tcb['pid']}\n")
    group = tcb.group

    if not group:
        return

    if group in processed_groups:
        return

    processed_groups.add(group)

    headers = ["FD", "FILEP", "OFLAGS", "POS", "PATH", "BACKTRACE"]
    formatter = "{:<4}{:<12}{:<12}{:<10}{:<48}{:<50}"
    gdb.write(formatter.format(*headers) + "\n")

    fd_count = 0
    for fd, fdp, file in foreach_file(tcb):
        print_file_info(fd, fdp, file, formatter)
        fd_count += 1
    total_fd_count += fd_count

    gdb.write("\n")


def foreach_file(tcb: Tcb):
    """Iterate over all file descriptors in a tcb"""
    group = tcb.group
    fdlist = group.tg_fdlist
    fl_fds = fdlist.fl_fds
    fl_rows = fdlist.fl_rows

    for row in range(fl_rows):
        for col in range(CONFIG_NFILE_DESCRIPTORS_PER_BLOCK):
            fdp = fl_fds[row][col]

            if not fdp or not fdp.f_file:
                continue

            fd = row * CONFIG_NFILE_DESCRIPTORS_PER_BLOCK + col
            file = fdp.f_file

            yield fd, fdp, file



############################################################################
# tools/pynuttx/rpmsgtrace2pcap.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import re
import struct
import sys
from dataclasses import dataclass
from typing import List, Optional

from nxelf.elf import ELFParser
from nxtrace.trace import NoteFactory, NoteParser, NotePrintf

RPMSG_NAME_SIZE = 32
RPMSG_TRANSPORT_VIRTIO = 1
RPMSG_TRANSPORT_PORT = 2


class tracePcap:
    def __init__(self):
        """Initialize the tracePcap with an empty packet list and current base time."""
        self.packets = []

    def process_input_trace(self, ts_sec, ts_usec, payload) -> None:
        self.packets.append((ts_sec, ts_usec, payload))

    def write_pcap(self, filename: str) -> None:
        """Write captured packets to a PCAP file in standard format."""
        with open(filename, "wb") as f:
            # PCAP Global Header
            f.write(
                struct.pack(
                    "<IHHiIII",
                    0xA1B2C3D4,  # Magic number (little-endian)
                    2,  # Version major
                    4,  # Version minor
                    0,  # Timezone (GMT/UTC)
                    0,  # Sigfigs
                    65535,  # Max packet length (snaplen)
                    147,  # Link-layer header type (DLT_USER_0, custom for RPMSG)
                )
            )

            # Write each packet with its header and data
            for ts_sec, ts_usec, packet_data in self.packets:
                # PCAP Packet Header
                f.write(
                    struct.pack(
                        "<IIII",
                        ts_sec,  # Timestamp seconds
                        ts_usec,  # Timestamp microseconds
                        len(packet_data),  # Length of packet on wire
                        len(packet_data),  # Length of packet stored in file
                    )
                )
                # Packet payload
                f.write(packet_data)


@dataclass
class Result:
    """Result structure containing parsed message information with optional header fields."""

    ts_sec: int
    ts_usec: int
    type: str
    name: str
    src: Optional[int] = None
    dst: Optional[int] = None
    service_type: Optional[int] = None
    length: Optional[int] = None
    flag: Optional[int] = None
    rdev: Optional[int] = None


def parse_printf(log_line: str, result: Result) -> Optional[Result]:
    """Parse RPMSG log line and extract relevant information."""

    # Determine message type
    if "get tx buffer" in log_line:
        result.type = "get_tx"
    elif "send ept" in log_line:
        result.type = "send_tx"
    elif "rx ept->cb start" in log_line:
        result.type = "rx_start"
    elif "rx ept->cb end" in log_line:
        result.type = "rx_end"
    elif "release tx buffer" in log_line:
        result.type = "release_tx"
    else:
        return None  # Unknown type

    # Extract name
    match = re.search(r"name:(\S+)", log_line)
    result.name = match.group(1) if match else "unknown"

    # Extract rdev (void* pointer) from all message types
    match = re.search(r"rdev:(\S+)", log_line)
    if match:
        rdev_str = match.group(1)
        result.rdev = int(rdev_str, 16)
    else:
        result.rdev = 0
        print(f"Warning: rdev not found in line: {log_line}")

    # Extract header information for relevant message types
    if result.type in ["send_tx", "rx_start"]:
        # Extract src and dst
        match = re.search(r"src:(\d+)", log_line)
        result.src = int(match.group(1)) if match else 0
        match = re.search(r"dst:(\d+)", log_line)
        result.dst = int(match.group(1)) if match else 0
        match = re.search(r"len:(\d+)", log_line)
        result.length = int(match.group(1)) if match else 0

        # Special handling for rpmsg-ns
        if result.src == 53 or result.dst == 53:
            result.name = "rpmsg-ns"

    return result


def parse_note_content(elf_parser, note) -> Optional[Result]:
    """
    Parse note content into a formatted string and timestamp in seconds.
    """
    result = Result(
        ts_sec=0,
        ts_usec=0,
        type="",
        name="",
        src=None,
        dst=None,
        length=None,
        flag=0,
        rdev=None,
    )

    time_ns = NoteFactory.cpu_cycles_to_ns(note.nc_systime)
    timestamp_s = time_ns / 1_000_000_000.0
    result.ts_sec = int(timestamp_s)
    result.ts_usec = int((timestamp_s - result.ts_sec) * 1000000)

    line = elf_parser.readstring(note.npt_fmt)
    if len(note.npt_data) > 0:
        line = NotePrintf().printf(elf_parser, line, note.npt_data)

    return parse_printf(line, result)


def build_message(result: Result, data: bytes) -> Optional[bytes]:
    """Build complete message based on result type.

    For 'send_tx' type: Merge header and data
    For 'rx_start' type: Return data only
    For other types: Return None

    Args:
        result: Result object containing header information
        data: Raw data bytes

    Returns:
        bytes: Complete message or None if type not supported
    """
    if result.type == "send_tx":
        # Convert Hdr structure to binary data
        # Format: src(4 bytes), dst(4 bytes), service_type(4 bytes), length(2 bytes), flag(2 bytes)
        src = result.src if result.src is not None else 0
        dst = result.dst if result.dst is not None else 0
        service_type = result.service_type if result.service_type is not None else 0
        length = result.length if result.length is not None else 0
        flag = result.flag if result.flag is not None else 0

        # Use little-endian byte order, format: I(4 bytes) I(4 bytes) I(4 bytes) H(2 bytes) H(2 bytes)
        hdr_binary = struct.pack("<3I2H", src, dst, service_type, length, flag)
        return hdr_binary + data
    elif result.type == "rx_start":
        return data
    else:
        return None


@dataclass
class ProtocolTable:
    """Protocol table structure for different RPMSG services."""

    name_pattern: str  # Name pattern supporting regular expressions
    level: int
    cmd_offset: int  # Command field offset in data
    cmd_len: int  # Command field length
    service_type: int  # Service type


# Create protocol table array
protocol_table = [
    ProtocolTable("rpmsg-ping", 1, 0, 0, 1),
    ProtocolTable("rpmsg-ns", 1, 0, 0, 2),
    ProtocolTable("s:.*", 2, 0, 4, 2),
    ProtocolTable("rpmsg-test", 1, 0, 0, 5),
    ProtocolTable("rpmsgfs-.*", 2, 0, 4, 5),
]


def update_message(rdev: int, name: str, message: bytes) -> bytes:
    """Update service type in message."""
    # Assume rpmsg_hdr length is 16 bytes
    rpmsg_hdr_length = 16
    SERVICE_TYPE_OFFSET = 8
    SERVICE_TYPE_LENGTH = 4

    # Check if message is valid
    if len(message) < 16:
        raise ValueError("Invalid message")

    """Merge transport type with service message."""
    transport = struct.pack("<I", rdev)

    for protocol in protocol_table:
        # Use regular expression to match name
        if re.match(protocol.name_pattern, name):
            if protocol.level == 1:
                # level=1, Update service type
                message = modify_binary_data_bytes(
                    message,
                    SERVICE_TYPE_OFFSET,
                    SERVICE_TYPE_LENGTH,
                    protocol.service_type,
                )
            elif protocol.level == 2:
                # level=2, extract command value from binary data
                # Calculate command data start position
                cmd_start = rpmsg_hdr_length + protocol.cmd_offset
                cmd_end = cmd_start + protocol.cmd_len

                if len(message) < cmd_end:
                    print(
                        f"Error: Binary data length insufficient, need at least \
                        {cmd_end} bytes"
                    )
                    return transport + message

                # Extract command data
                cmd_value = int.from_bytes(
                    message[cmd_start:cmd_end], byteorder="little"
                )

                # Return message
                message = modify_binary_data_bytes(
                    message,
                    SERVICE_TYPE_OFFSET,
                    SERVICE_TYPE_LENGTH,
                    protocol.service_type + cmd_value,
                )

            return transport + message

    return transport + message


def modify_binary_data_bytes(
    original_data: bytes, offset: int, length: int, new_value: int
) -> bytes:
    """Modify binary data at specified offset and length with new value."""
    # Check if offset and length are valid
    if offset < 0 or offset + length > len(original_data):
        raise ValueError(
            f"Invalid offset or length: offset={offset}, length={length}, data length={len(original_data)}"
        )

    # Convert new value to byte array of specified length (little-endian)
    new_bytes = new_value.to_bytes(length, byteorder="little")

    # Create new byte array and replace data
    modified_data = bytearray(original_data)
    modified_data[offset : offset + length] = new_bytes

    return bytes(modified_data)


def note_binary_parser(elf_parser, note_parser, freq) -> List[str]:
    """Parse binary notes and process RPMSG messages."""
    # Initialize NoteFactory
    NoteFactory.init_instance(
        elf_parser=elf_parser, output="output_trace.perfetto", frequency_hz=freq
    )

    result = None
    message = None
    pcap_processor = tracePcap()
    note_type = elf_parser.get_type("note_type_e")

    for note in note_parser.notes:
        # Step 1: Parse text content
        if note.nc_type == note_type.NOTE_DUMP_PRINTF:
            if message is not None:
                message = update_message(result.rdev, result.name, message)
                pcap_processor.process_input_trace(
                    result.ts_sec, result.ts_usec, message
                )
            message = None
            result = parse_note_content(elf_parser, note)
        elif note.nc_type == note_type.NOTE_DUMP_BINARY and result is not None:
            if message is None:
                message = build_message(result, note.nev_data)
            else:
                message = message + note.nev_data

    pcap_processor.write_pcap("rpmsg_trace.pcap")


def arg_parse():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--elf", help="ELF file", required=True)
    parser.add_argument("-o", "--output", help="Output file", default="trace.perfetto")
    parser.add_argument("-b", "--binary", help="Binary note file")
    parser.add_argument(
        "-f", "--freq", type=int, default=1000000000, help="Frequency of the trace"
    )
    args = parser.parse_args()

    # Validate the arguments
    if args.binary:
        return vars(args)
    elif args.device and args.interface:
        return vars(args)
    elif args.device or args.interface:
        print("Error: Both --device and --interface must be specified together.")
        sys.exit(1)
    else:
        print(
            "Error: You must specify either a binary file or both RTT parameters (--device and --interface)."
        )
        sys.exit(1)


if __name__ == "__main__":
    args = arg_parse()
    elf_parser = ELFParser(args["elf"])
    note_parser = NoteParser(elf_parser, output=args["output"])
    note_parser.parse_file(args["binary"])
    note_parser.dump()
    note_binary_parser(elf_parser, note_parser, args["freq"])



############################################################################
# tools/pynuttx/nxgdb/fs/utils.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from .. import utils

FSNODEFLAG_TYPE_MASK = 0x0000000F

CONFIG_PSEUDOFS_FILE = utils.lookup_type("struct fs_pseudofile_s") is not None
CONFIG_PSEUDOFS_ATTRIBUTES = utils.has_field("struct inode", "i_mode")

CONFIG_FS_BACKTRACE = utils.has_field("struct fd", "f_backtrace")
CONFIG_FS_SHMFS = utils.lookup_type("struct shmfs_object_s") is not None

CONFIG_NFILE_DESCRIPTORS_PER_BLOCK = utils.get_field_nitems(
    "struct fdlist", "fl_prefds"
)

# see fs_mount.c
CONFIG_DISABLE_MOUNTPOINT = utils.lookup_type("struct fsmap_t") is None

# see fs_romfs.c
CONFIG_FS_ROMFS_CACHE_NODE = utils.has_field("struct romfs_dir_s", "firstnode")

# see lfs_vfs.c
CONFIG_FS_LITTLEFS = utils.lookup_type("struct littlefs_file_s") is not None

# see yaffs_vfs.c
CONFIG_FS_YAFFS = utils.lookup_type("struct yaffs_file_s") is not None

# see fatfs_vfs.c
CONFIG_FS_FATFS = utils.lookup_type("struct fatfs_file_s") is not None

# see fs_Irofs.c
CONFIG_FS_LROFS = utils.lookup_type("struct lrofs_dir_s") is not None



############################################################################
# tools/pynuttx/gmcp/nxgdbmcp/gmcp/tools/session.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import uuid
from typing import Dict, List, Optional

from mcp.server.fastmcp import Context

from ..context import get_session
from ..session.gdb_session import GdbSession
from ..utils import _exec_command, error_handler, get_available_port


@error_handler
async def _start(host: Optional[str] = None, port: Optional[int] = None) -> GdbSession:
    session_id = str(uuid.uuid4())

    session = GdbSession(session_id, host, port)

    return session


@error_handler
async def _load_program(
    ctx: Context,
    session_id: str,
    program: str,
    arguments: Optional[List[str]] = None,
) -> str:
    session = get_session(ctx, session_id)

    # Load executable
    output = await session.execute_command(f"file {program}")

    # Set program arguments if provided
    if arguments:
        args_str = " ".join(f"{arg}" for arg in arguments)
        args_output = await session.execute_command(f"set args {args_str}")
        output += f"\n{args_output}"

    return f"Program loaded: {program}\n\nOutput:\n{output}"


@error_handler
async def _load_program_with_parsed_core(
    ctx: Context,
    session_id: str,
    program: str,
    core_path: str,
    nxgdb_path: str = "nuttx/tools/pynuttx/gdbinit.py",
) -> str:
    session = get_session(ctx, session_id)

    # First load the program
    file_output = await session.execute_command(f"file {program}")

    await session.execute_command(f"source {nxgdb_path}")

    # Find an available port for the stub
    port = get_available_port()
    # Then load the core file with the port
    core_output = await session.execute_command(
        f"target nxstub --core {core_path} --port {port}"
    )

    return f"Core file loaded: {core_path} \n\nOutput:\n{file_output}\n{core_output}"


@error_handler
async def _load_program_with_parsed_memdump(
    ctx: Context,
    session_id: str,
    program: str,
    rawfiles: list[str],
    nxgdb_path: str = "nuttx/tools/pynuttx/gdbinit.py",
) -> str:
    session = get_session(ctx, session_id)

    # First load the program
    file_output = await session.execute_command(f"file {program}")

    await session.execute_command(f"source {nxgdb_path}")

    # Find an available port for the stub
    port = get_available_port()
    # Prepare rawfile arguments string
    rawfile_args = " ".join(rawfiles)
    # Then load the memory dump(s) with the port
    core_output = await session.execute_command(
        f"target nxstub --rawfile {rawfile_args} --port {port}"
    )

    return f"Memory dump(s) loaded: {rawfile_args} \n\nOutput:\n{file_output}\n{core_output}"


def register_session_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def gdb_connect(
        ctx: Context, host: Optional[str] = None, port: Optional[int] = None
    ) -> str:
        """Establish a connection with GDB via socket with optional host and port.
        This action will not load coredump or do anything else.

        Args:
            ctx (Context): The context object containing runtime/session information.
            host (Optional[str]): Host IP for the GDB debugging session, default to localhost if None.
            port (Optional[int]): Port number for the GDB debugging session, default to 20819 if None.

        Returns:
            session (GdbSession): An instance of GdbSession representing the established connection.
        """
        session = await _start(host, port)

        ctx.request_context.lifespan_context.sessions[session.id] = session
        return f"GDB session connected, ID: {session.id}\n"

    @gdb_mcp.tool()
    async def gdb_target_remote(ctx: Context, session_id: str, remote: str) -> str:
        """Connect to a remote gdbserver via target remote ip:port"""
        return await _exec_command(ctx, session_id, f"target remote {remote}")

    @gdb_mcp.tool()
    async def gdb_start_with_remote(
        ctx: Context,
        program: str,
        remote: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> str:
        """Establish a connection with GDB via socket with optional host and port,
        load file, and connect to remote target in one step."""
        # Start GDB session first
        session = await _start(host, port)

        output1 = await session.execute_command(f"file {program}")
        output2 = await session.execute_command(f"target remote {remote}")

        ctx.request_context.lifespan_context.sessions[session.id] = session
        return f"GDB session connected, ID: {session.id}\n\nFile output:\n{output1}\n\nRemote output:\n{output2}"

    @gdb_mcp.tool()
    async def gdb_source_python(ctx: Context, session_id: str, script_path: str) -> str:
        """Source a python script in GDB (using 'source <script>')"""
        return await _exec_command(ctx, session_id, f"source {script_path}")

    @gdb_mcp.tool()
    async def gdb_load_program(
        ctx: Context,
        session_id: str,
        program: str,
        arguments: Optional[List[str]] = None,
    ) -> str:
        """Load a program into GDB without coredump"""
        return await _load_program(ctx, session_id, program, arguments)

    @gdb_mcp.tool()
    async def gdb_load_program_with_core(
        ctx: Context, session_id: str, program: str, core_path: str
    ) -> str:
        """Load program with a core dump file"""
        output = await _exec_command(ctx, session_id, f"file {program}")
        output += await _exec_command(ctx, session_id, f"core {core_path}")
        return output

    @gdb_mcp.tool()
    async def gdb_load_program_with_parsed_core(
        ctx: Context,
        session_id: str,
        program: str,
        core_path: str,
        nxgdb_path: str = "nuttx/tools/pynuttx/gdbinit.py",
    ) -> str:
        """
        Load a program and parse the coredump using nxgdb, so GDB can debug with the NuttX thread model.

        This function first loads the executable with the 'file' command,
        then sources nxgdb(e.g., nuttx/tools/pynuttx/gdbinit.py),
        and finally loads the coredump with 'target nxstub --core <core> --port <port>'.
        This allows GDB to correctly recognize NuttX threads and debugging context,
        greatly improving the debugging experience.
        Note: nxgdb must be loaded before loading the coredump. The port is dynamically selected.
        """
        return await _load_program_with_parsed_core(
            ctx, session_id, program, core_path, nxgdb_path
        )

    @gdb_mcp.tool()
    async def gdb_load_program_with_parsed_memdump(
        ctx: Context,
        session_id: str,
        program: str,
        rawfiles: List[str],
        nxgdb_path: str = "nuttx/tools/pynuttx/gdbinit.py",
    ) -> str:
        """
        Load a program and parse one or more memory dump files using nxgdb, so GDB can debug with the NuttX thread model.

        Args:
            ctx: Context object
            session_id: GDB session ID
            program: Path to the program executable
            rawfiles: List of memory dump file arguments, each in the format 'memdump.bin:address'
            nxgdb_path: Path to nxgdb Python script (default: nuttx/tools/pynuttx/gdbinit.py)

        This function first loads the executable with the 'file' command,
        then sources nxgdb (e.g., nuttx/tools/pynuttx/gdbinit.py),
        and finally loads the memory dump(s) with
        'target nxstub --rawfile <file1:addr1> <file2:addr2> ... --port <port>'.
        This allows GDB to correctly recognize NuttX threads and debugging context,
        greatly improving the debugging experience.
        Note: nxgdb must be loaded before loading the memory dump. The port is dynamically selected.
        """
        return await _load_program_with_parsed_memdump(
            ctx, session_id, program, rawfiles, nxgdb_path
        )

    @gdb_mcp.tool()
    async def gdb_attach(ctx: Context, session_id: str, pid: int) -> str:
        """Attach to a running process"""
        return await _exec_command(ctx, session_id, f"attach {pid}")

    @gdb_mcp.tool()
    async def gdb_terminate(ctx: Context, session_id: str) -> str:
        """Terminate a GDB session"""
        try:
            session = get_session(ctx, session_id)
            await session.cleanup()
            ctx.request_context.lifespan_context.sessions.pop(session_id, None)
            return f"GDB session terminated: {session_id}"

        except ValueError as e:
            return str(e)
        except Exception as e:
            return f"Failed to terminate GDB session: {str(e)}"

    @gdb_mcp.tool()
    def gdb_list_sessions(ctx: Context) -> str:
        """List all active GDB sessions"""
        sessions: Dict[str, GdbSession] = ctx.request_context.lifespan_context.sessions
        session_info = []

        for session_id, session in sessions.items():
            session_info.append(
                {
                    "id": session_id,
                    "host": session.host,
                    "port": session.port,
                }
            )

        return f"Active GDB Sessions ({len(sessions)}):\n\n{session_info}"

    @gdb_mcp.tool()
    async def gdb_run(ctx: Context, session_id: str) -> str:
        """Run the loaded program"""
        return await _exec_command(ctx, session_id, "run")

    @gdb_mcp.tool()
    async def gdb_kill(ctx: Context, session_id: str) -> str:
        """Kill the running process"""
        return await _exec_command(ctx, session_id, "kill")



############################################################################
# tools/pynuttx/nxgdb/mm.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

import argparse
from typing import Generator, List, Tuple

import gdb

from . import autocompeletion, backtrace, lists, utils
from .backtrace import CONFIG_LIBC_BACKTRACE_DEPTH
from .protocols import mm as p
from .utils import Value

# Note we use mm_freenode_s to check if CONFIG_MM_RECORD_STACK is enabled instead
# of utils.get_symbol_value("CONFIG_MM_RECORD_STACK") because the latter may report
# wrong value on some platforms.

CONFIG_MM_RECORD_STACK = utils.has_field("struct mm_freenode_s", "stack")
CONFIG_MM_RECORD_PID = utils.has_field("struct mm_freenode_s", "pid")
CONFIG_MM_RECORD_SEQNO = utils.has_field("struct mm_freenode_s", "seqno")
CONFIG_MM_RECORD = (
    CONFIG_MM_RECORD_STACK or CONFIG_MM_RECORD_PID or CONFIG_MM_RECORD_SEQNO
)
MM_RECORD_STACK_DEPTH = CONFIG_LIBC_BACKTRACE_DEPTH if CONFIG_MM_RECORD_STACK else 0

mempool_record_s = utils.lookup_type("struct mempool_record_s")
mm_record_size = 0

g_sections = None

PID_MM_INVALID = -100
PID_MM_MEMPOOL = -1


def get_sections():
    global g_sections
    if g_sections is None:
        g_sections = gdb.execute("maintenance info sections", to_string=True)
    return g_sections


def mm_alignup(size: int) -> int:
    size_t = utils.lookup_type("uintptr_t")
    CONFIG_MM_DEFAULT_ALIGNMENT = 2 * size_t.sizeof
    align = CONFIG_MM_DEFAULT_ALIGNMENT
    # @todo: use the actual value of CONFIG_MM_DEFAULT_ALIGNMENT
    # align = utils.get_symbol_value("CONFIG_MM_DEFAULT_ALIGNMENT") or 2 * size_t.sizeof
    size = (size + align - 1) & ~(align - 1)
    return size


if mempool_record_s and CONFIG_MM_RECORD:
    mm_record_size = mm_alignup(mempool_record_s.sizeof)


class MemPoolBlock:
    """
    Memory pool block instance.
    """

    MAGIC_ALLOC = 0xAAAA_AAAA

    def __init__(
        self, addr: int, blocksize: int, overhead: int, pool: MemPool = None
    ) -> None:
        """
        Initialize the memory pool block instance.
        block: must be start address of the block,
        blocksize: block size without backtrace overhead,
        overhead: backtrace overhead size.
        pool: the MemPool instance that this block belongs to.
        """
        self.overhead = overhead
        self.from_pool = True
        self.is_orphan = False
        self.address = addr
        self.blocksize = int(blocksize)
        self.nodesize = int(blocksize) + self.overhead
        self.usersize = self.blocksize
        self.useraddress = self.address
        self.pool = pool
        # Lazy evaluation
        self._backtrace = self._pid = self._seqno = self._magic = self._record = None

    def __repr__(self) -> str:
        return f"block@{hex(self.address)},size:{self.blocksize},seqno:{self.seqno},pid:{self.pid}"

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        return hash((self.pid, self.nodesize, self.backtrace))

    def __eq__(self, value: MemPoolBlock) -> bool:
        return (
            self.pid == value.pid
            and self.nodesize == value.nodesize
            and self.backtrace == value.backtrace
        )

    def contains(self, address: int) -> bool:
        """Check if the address is in block's range, excluding overhead"""
        return self.address <= address < self.address + self.blocksize

    @property
    def record(self) -> p.MemPoolBlock:
        if not self._record:
            addr = int(self.address)
            addr -= mm_record_size
            self._record = (
                gdb.Value(addr).cast(mempool_record_s.pointer()).dereference()
            )
        return self._record

    @property
    def is_free(self) -> bool:
        if not CONFIG_MM_RECORD:
            return self.pool.is_free(self) if self.pool else None

        if not self._magic:
            self._magic = int(self.record["magic"])

        return self._magic != self.MAGIC_ALLOC

    @property
    def seqno(self) -> int:
        if not self._seqno:
            self._seqno = (
                int(self.record["seqno"]) if CONFIG_MM_RECORD_SEQNO else PID_MM_INVALID
            )
        return self._seqno

    @property
    def pid(self) -> int:
        if not self._pid:
            self._pid = (
                int(self.record["pid"]) if CONFIG_MM_RECORD_PID else PID_MM_INVALID
            )
        return self._pid

    @property
    def backtrace(self) -> Tuple[int]:
        if MM_RECORD_STACK_DEPTH <= 0:
            return ()

        if not self._backtrace:
            self._backtrace = tuple(
                backtrace.BacktraceEntry(self.record["stack"]).get()
            )
        return self._backtrace

    @property
    def prevnode(self) -> MemPoolBlock:
        addr = self.address - self.nodesize
        return MemPoolBlock(addr, self.blocksize, self.overhead, pool=self.pool)

    @property
    def nextnode(self) -> MemPoolBlock:
        addr = self.address + self.nodesize
        return MemPoolBlock(addr, self.blocksize, self.overhead, pool=self.pool)

    def read_memory(self) -> memoryview:
        return gdb.selected_inferior().read_memory(self.address, self.blocksize)


class MemPool(Value, p.MemPool):
    """
    Memory pool instance.
    """

    def __init__(self, mpool: Value, name=None) -> None:
        if mpool.type.code == gdb.TYPE_CODE_PTR:
            mpool = mpool.dereference()
        super().__init__(mpool)
        self._blksize = None
        self._nfree = None
        self._nifree = None
        self._overhead = None
        self._free_blks = None

    def __repr__(self) -> str:
        return f"{self.name}@{hex(self.address)},size:{self.size}/{self['blocksize']},nused:{self.nused},nfree:{self.nfree}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def name(self) -> str:
        try:
            return self.procfs.name.string()
        except Exception:
            return "<noname>"

    @property
    def memranges(self) -> Generator[Tuple[int, int], None, None]:
        """Memory ranges of the pool"""
        sq_entry_t = utils.lookup_type("sq_entry_t")
        blksize = self.size

        if self.ibase:
            blks = int(self.interruptsize) // blksize
            base = int(self.ibase)
            yield (base, base + blks * blksize)

        if not self.equeue.head:
            return None

        # First queue has size of initialsize
        ninit = int(self.initialsize)
        ninit = ninit and (ninit - sq_entry_t.sizeof) // blksize
        nexpand = (int(self.expandsize) - sq_entry_t.sizeof) // blksize

        for entry in lists.NxSQueue(self.equeue):
            blks = ninit or nexpand
            ninit = 0
            yield (int(entry) - blks * blksize, int(entry))

    @property
    def nqueue(self) -> int:
        return lists.sq_count(self.equeue)

    @property
    def size(self) -> int:
        """Real block size including backtrace overhead"""
        if not self._blksize:
            blksize = self["blocksize"]
            if CONFIG_MM_RECORD:
                blksize = mm_alignup(blksize + mempool_record_s.sizeof)
            self._blksize = int(blksize)
        return self._blksize

    @property
    def overhead(self) -> int:
        if not self._overhead:
            self._overhead = self.size - int(self["blocksize"])
        return self._overhead

    @property
    def nwaiter(self) -> int:
        return (
            -int(self.waitsem.val.semcount) if self.wait and self.expandsize == 0 else 0
        )

    @property
    def nused(self) -> int:
        return int(self.nalloc)

    @property
    def free(self) -> int:
        return (self.nfree + self.nifree) * self.size

    @property
    def nfree(self) -> int:
        if not self._nfree:
            self._nfree = lists.sq_count(self.queue)
        return self._nfree + self.nifree

    @property
    def nifree(self) -> int:
        """Interrupt pool free blocks count"""
        if not self._nifree:
            self._nifree = lists.sq_count(self.iqueue)
        return self._nifree

    @property
    def total(self) -> int:
        nqueue = lists.sq_count(self.equeue)
        sq_entry_t = utils.lookup_type("sq_entry_t")
        blocks = self.nused + self.nfree
        return int(nqueue * sq_entry_t.sizeof + blocks * self.size)

    @property
    def blks(self) -> Generator[MemPoolBlock, None, None]:
        """Iterate over all blocks in the pool"""
        sq_entry_t = utils.lookup_type("sq_entry_t")
        blksize = self.size  # Real block size including backtrace overhead
        blocksize = self["blocksize"]

        def iterate(entry, nblocks):
            base = int(entry) - nblocks * blksize
            while nblocks > 0:
                yield MemPoolBlock(
                    base + mm_record_size, blocksize, self.overhead, pool=self
                )
                base += blksize
                nblocks -= 1

        if self.ibase:
            blks = int(self.interruptsize) // blksize
            yield from iterate(self.ibase + blks * blksize, blks)

        if not self.equeue.head:
            return None

        # First queue has size of initialsize
        ninit = int(self.initialsize)
        ninit = ninit and (ninit - sq_entry_t.sizeof) // blksize
        nexpand = (int(self.expandsize) - sq_entry_t.sizeof) // blksize

        for entry in lists.NxSQueue(self.equeue):
            yield from iterate(entry, ninit or nexpand)
            ninit = 0

    def contains(self, address: int) -> Tuple[bool, Value]:
        ranges = self.memranges
        if not ranges:
            return False, None

        for start, end in ranges:
            if start <= address < end:
                return True, None

    def find(self, address: int) -> Value:
        """Find the block that contains the given address"""
        sq_entry_t = utils.lookup_type("sq_entry_t")
        blksize = self.size
        blocksize = self["blocksize"]

        def get_blk(base):
            blkstart = base + (address - base) // blksize * blksize
            blkstart += mm_record_size
            return MemPoolBlock(blkstart, blocksize, self.overhead, pool=self)

        if self.ibase:
            # Check if it belongs to interrupt pool
            blks = int(self.interruptsize) // blksize
            base = int(self.ibase)
            if base <= address < base + blks * blksize:
                return get_blk(base)

        if not self.equeue.head:
            return None

        # First queue has size of initialsize
        ninit = int(self.initialsize)
        ninit = ninit and (ninit - sq_entry_t.sizeof) // blksize
        nexpand = (int(self.expandsize) - sq_entry_t.sizeof) // blksize

        for entry in lists.NxSQueue(self.equeue):
            blks = ninit or nexpand
            ninit = 0
            base = int(entry) - blks * blksize
            if base <= address < int(entry):
                return get_blk(base)

    def blks_free(self) -> Generator[MemPoolBlock, None, None]:
        """Iterate over all free blocks in the pool"""
        blocksize = self["blocksize"]
        for entry in lists.NxSQueue(self.queue):
            yield MemPoolBlock(int(entry), blocksize, self.overhead, pool=self)

    def blks_used(self) -> Generator[MemPoolBlock, None, None]:
        """Iterate over all used blocks in the pool"""
        return filter(lambda blk: not blk.is_free, self.blks)

    def is_free(self, blk: MemPoolBlock) -> bool:
        """Check if the given block is free in the pool"""
        if self._free_blks is None:
            try:
                self._free_blks = set(free.address for free in self.blks_free())
            except Exception:
                self._free_blks = set()

            # monitor GDB stop event to clear cache
            def clear_free_blks(event):
                self._free_blks = None

            gdb.events.stop.connect(clear_free_blks)

        # If self._free_blks is empty, it's high possible there's problem reading free blocks
        # instead of all memory is used.
        if not self._free_blks:
            return None

        return blk.address in self._free_blks


class MemPoolMultiple(Value, p.MemPoolMultiple):
    """
    Multiple level memory pool instance.
    """

    def __init__(self, mpool: Value, name=None) -> None:
        if mpool.type.code == gdb.TYPE_CODE_PTR:
            mpool = mpool.dereference()
        super().__init__(mpool)

    def __repr__(self) -> str:
        return f"Multiple Level Memory Pool: {self.address}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def pools(self) -> Generator[MemPool, None, None]:
        for pool in utils.ArrayIterator(self["pools"], self.npools):
            yield MemPool(pool)

    @property
    def free(self) -> int:
        return sum(pool.free for pool in self.pools)

    @property
    def chunks(self) -> Generator[MemPoolBlock, None, None]:
        for chunk in lists.sq_for_every(self.chunk_queue):
            chunk = chunk.cast(gdb.lookup_type("struct mpool_chunk_s"))
            yield chunk


class MMNode(gdb.Value, p.MMFreeNode):
    """
    One memory node in the memory manager heap, either free or allocated.
    The instance is always dereferenced to the actual node.
    """

    MM_ALLOC_BIT = 0x1
    MM_PREVFREE_BIT = 0x2
    MM_MASK_BIT = MM_ALLOC_BIT | MM_PREVFREE_BIT
    try:
        MM_SIZEOF_ALLOCNODE = utils.sizeof("struct mm_allocnode_s")
        # Although preceding can locates in the previous node, we still count it as overhead
        MM_ALLOCNODE_OVERHEAD = MM_SIZEOF_ALLOCNODE
        MM_MIN_SHIFT = utils.log2ceil(utils.sizeof("struct mm_freenode_s"))
        MM_MIN_CHUNK = 1 << MM_MIN_SHIFT
    except Exception:
        MM_SIZEOF_ALLOCNODE = 0
        MM_ALLOCNODE_OVERHEAD = 0
        MM_MIN_SHIFT = 0
        MM_MIN_CHUNK = 0

    def __init__(self, node: gdb.Value):
        if node.type.code == gdb.TYPE_CODE_PTR:
            node = node.dereference()
        self._backtrace = None
        self._address = None
        self._nodesize = None
        super().__init__(node)

    def __repr__(self):
        return (
            f"{hex(self.address)}({'F' if self.is_free else 'A'}{'F' if self.is_prev_free else 'A'})"
            f" size:{self.nodesize}/{self.prevsize if self.is_prev_free else '-'}"
            f" seq:{self.seqno} pid:{self.pid} "
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        return hash((self.pid, self.nodesize, self.backtrace))

    def __eq__(self, value: MMNode) -> bool:
        return (
            self.pid == value.pid
            and self.nodesize == value.nodesize
            and self.backtrace == value.backtrace
        )

    def contains(self, address):
        """Check if the address is in node's range, excluding overhead"""
        return self.useraddress <= address < self.useraddress + self.usersize

    def read_memory(self):
        return gdb.selected_inferior().read_memory(self.useraddress, self.usersize)

    @property
    def address(self) -> int:
        """Change 'void *' to int"""
        if not self._address:
            self._address = int(super().address)
        return self._address

    @property
    def useraddress(self) -> int:
        """Address of user memory, excluding overhead"""
        return self.address + self.overhead

    @property
    def prevsize(self) -> int:
        """Size of preceding chunk size"""
        return int(self["preceding"]) & ~MMNode.MM_MASK_BIT

    @property
    def nodesize(self) -> int:
        """Size of this chunk, including overhead"""
        if not self._nodesize:
            self._nodesize = int(self["size"]) & ~MMNode.MM_MASK_BIT
        return self._nodesize

    @property
    def usersize(self) -> int:
        """Size of this chunk, excluding overhead"""
        usersize = self.nodesize - MMNode.MM_ALLOCNODE_OVERHEAD

        # This node is allocted, thus the next node's preceding is user memory.
        usersize += 4
        return usersize

    @property
    def flink(self):
        # Only free node has flink and blink
        return MMNode(self["flink"]) if self.is_free and self["flink"] else None

    @property
    def blink(self):
        # Only free node has flink and blink
        return MMNode(self["blink"]) if self.is_free and self["blink"] else None

    @property
    def pid(self) -> int:
        if CONFIG_MM_RECORD_PID:
            return int(self["pid"])
        return PID_MM_INVALID

    @property
    def seqno(self) -> int:
        return int(self["seqno"]) if CONFIG_MM_RECORD_SEQNO else -1

    @property
    def backtrace(self) -> List[Tuple[int, str, str]]:
        if MM_RECORD_STACK_DEPTH <= 0:
            return ()

        # The free mm heap node does not record backtrace, the value may be illegal
        try:
            if not self._backtrace and (
                stack := backtrace.BacktraceEntry(self["stack"]).get()
            ):
                self._backtrace = tuple(stack)
        except gdb.MemoryError:
            self._backtrace = tuple(backtrace.BacktraceEntry(0).get())

        return self._backtrace

    @property
    def prevnode(self) -> MMNode:
        if not self.is_prev_free:
            return None

        addr = int(self.address) - self.prevsize
        type = utils.lookup_type("struct mm_freenode_s").pointer()
        return MMNode(gdb.Value(addr).cast(type))

    @property
    def nextnode(self) -> MMNode:
        if not self.nodesize:
            gdb.write(f"\n\x1b[31;1m Node corrupted: {self} \x1b[m\n")
            return None

        addr = int(self.address) + self.nodesize
        type = utils.lookup_type("struct mm_freenode_s").pointer()
        # Use gdb.Value for better performance
        return MMNode(gdb.Value(addr).cast(type))

    @property
    def is_free(self) -> bool:
        return not self["size"] & MMNode.MM_ALLOC_BIT

    @property
    def is_prev_free(self) -> bool:
        return self["size"] & MMNode.MM_PREVFREE_BIT

    @property
    def is_orphan(self) -> bool:
        # Report orphaned node and node likely to be orphaned(free-used-used-free)
        return self.is_prev_free or self.nextnode.is_free

    @property
    def from_pool(self) -> bool:
        return False

    @property
    def overhead(self) -> int:
        return MMNode.MM_ALLOCNODE_OVERHEAD


class MMHeap(Value, p.MMHeap):
    """
    One memory manager heap. It may contains multiple regions.
    """

    def __init__(self, heap: Value, name=None) -> None:
        mm_heap_s = utils.lookup_type("struct mm_heap_s")
        if isinstance(heap, int) or heap.type.code == gdb.TYPE_CODE_INT:
            heap = gdb.Value(heap).cast(mm_heap_s.pointer()).dereference()
        elif heap.type.code == gdb.TYPE_CODE_PTR:
            heap = heap.dereference()

        if heap.type != mm_heap_s:
            raise ValueError(f"Invalid heap type: {heap.type}")

        super().__init__(heap)

        self.name = name or "<noname>"
        self._regions = None

        # Check if heap node is accessible
        try:
            for start, end in self.regions:
                gdb.selected_inferior().read_memory(start.address, 1)
                gdb.selected_inferior().read_memory(end.address, 1)
        except gdb.MemoryError:
            raise ValueError(f"Heap node not accessible: {heap}")

    def __repr__(self) -> str:
        regions = [
            f"{hex(start.address)}~{hex(end.address)}" for start, end in self.regions
        ]
        return f"{self.name}@{self.address}, {int(self.heapsize) / 1024 :.1f}kB {self.nregions}regions: {','.join(regions)}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def curused(self) -> int:
        return int(self.mm_curused)

    @property
    def heapsize(self) -> int:
        return int(self.mm_heapsize)

    @property
    def free(self) -> int:
        return self.heapsize - self.curused

    @property
    def nregions(self) -> int:
        return int(utils.get_field(self, "mm_nregions", default=1))

    @property
    def mm_mpool(self) -> MemPoolMultiple:
        mpool = utils.get_field(self, "mm_mpool", default=None)
        return MemPoolMultiple(mpool) if mpool else None

    @property
    def regions(self) -> List[Tuple[MMNode, MMNode]]:
        if not self._regions:
            regions = self.nregions
            self._regions = []
            for start, end in zip(
                utils.ArrayIterator(self.mm_heapstart, regions),
                utils.ArrayIterator(self.mm_heapend, regions),
            ):
                self._regions.append((MMNode(start), MMNode(end)))
        return self._regions

    @property
    def nodes(self) -> Generator[MMNode, None, None]:
        for start, end in self.regions:
            node = start
            while node and node.address <= end.address:
                yield node
                node = node.nextnode

    def nodes_free(self) -> Generator[MMNode, None, None]:
        return filter(lambda node: node.is_free, self.nodes)

    def nodes_used(self) -> Generator[MMNode, None, None]:
        return filter(lambda node: not node.is_free, self.nodes)

    def contains(self, address: int) -> bool:
        ranges = [[int(start.address), int(end.address)] for start, end in self.regions]
        ranges[0][0] = int(self.address)  # The heap itself is also in the range
        return any(start <= address <= end for start, end in ranges)

    def find(self, address: int) -> MMNode:
        for node in self.nodes:
            if node.address <= address < node.address + node.nodesize:
                return node


def get_heaps() -> List[MMHeap]:
    # parse g_procfs_meminfo to get all heaps
    heaps = []
    meminfo: p.ProcfsMeminfoEntry = utils.gdb_eval_or_none("g_procfs_meminfo")
    if not meminfo and (heap := utils.parse_and_eval("g_mmheap")):
        try:
            heaps.append(MMHeap(heap))
        except Exception:
            pass

    try:
        while meminfo:
            try:
                heap = MMHeap(meminfo.heap, name=meminfo.name.string())
                heaps.append(heap)
            except Exception:
                pass

            meminfo = meminfo.next
    except gdb.MemoryError:
        # procfs not accessible
        pass

    return heaps


def get_pools(heaps: List[Value] = []) -> Generator[MemPool, None, None]:
    for heap in heaps or get_heaps():
        if not (mm_pool := heap.mm_mpool):
            continue

        for pool in mm_pool.pools:
            yield pool


def memory_range(heap=True, globals=True) -> List[Tuple[int, int]]:
    # Execute the GDB command to get section info
    memranges = []
    if globals:
        sections = get_sections()
        # Parse the output to find sections with ALLOC and LOAD
        for line in sections.splitlines():
            if "ALLOC" in line and "READONLY" not in line:
                parts = line.split()
                start = int(parts[1].split("->")[0], 16)
                end = int(parts[1].split("->")[1], 16)
                if start == end:
                    continue

                memranges.append((start, end))

        idle_topstack = int(utils.parse_and_eval("g_idle_topstack").cast("uintptr_t"))
        idle_stacksize = int(utils.parse_and_eval("CONFIG_IDLETHREAD_STACKSIZE"))
        memranges.append((idle_topstack - idle_stacksize, idle_topstack))

    # Get heaps from memdump
    if heap:
        for heap in get_heaps():
            for i in range(heap.nregions):
                start = int(heap["mm_heapstart"][i])
                end = int(heap["mm_heapend"][i]) + MMNode.MM_SIZEOF_ALLOCNODE

                if start == end:
                    continue

                # For the first region in the heap we need to compensate for the heap_s size
                if i == 0:
                    start = int(heap.address)

                for r in memranges:
                    # If the address range is already in memranges, skip
                    if r[0] <= start and r[1] >= end:
                        break
                    # If the new address range includes a range in memranges,
                    # delete the old one and add the new one
                    elif start <= r[0] and end >= r[1]:
                        memranges.remove(r)
                        memranges.append((start, end))
                        break
                else:
                    memranges.append((start, end))
    return sorted(memranges, key=lambda x: x[0])


def get_memrange(
    rangestr: str = "", heap_only=False, globals_only=False
) -> List[Tuple[int, int]]:
    """
    Parse memory range from string or get from heap and globals,
    the string should be in the format of "start1,end1,attr1 start2,end2,attr2".
    """

    memrange = []
    if rangestr:
        values = rangestr.replace('"', "").split(",")
        for i in range(0, len(values), 3):
            start = utils.parse_arg(values[i])
            end = utils.parse_arg(values[i + 1])
            memrange.append((start, end))
    else:
        memrange = memory_range(heap=not globals_only, globals=not heap_only)

    # Merge overlapping ranges
    merged_ranges = []
    for start, end in sorted(memrange):
        if merged_ranges and start <= merged_ranges[-1][1]:
            merged_ranges[-1] = (merged_ranges[-1][0], max(merged_ranges[-1][1], end))
        else:
            merged_ranges.append((start, end))

    return merged_ranges


def get_nodes_dict():
    """
    Return dict of all memory nodes, including memory pool.
    """

    # If there is a cached result, return it directly
    if hasattr(get_nodes_dict, "_cached_nodes"):
        return get_nodes_dict._cached_nodes

    nodes = []
    for heap in get_heaps():
        nodes.extend(
            {
                "name": heap.name,
                "address": node.address,
                "size": node.nodesize,
                "seqno": node.seqno,
                "pid": node.pid,
                "free": node.is_free,
                "from_pool": False,
                "backtrace": node.backtrace,
            }
            for node in heap.nodes
        )

        for pool in get_pools([heap]):
            nodes.extend(
                {
                    "name": f"{heap.name}@{blk.nodesize}pool",
                    "address": blk.address,
                    "size": blk.nodesize,
                    "seqno": blk.seqno,
                    "pid": blk.pid,
                    "free": blk.is_free,
                    "from_pool": True,
                    "backtrace": blk.backtrace,
                }
                for blk in pool.blks
            )

    # Cache the result
    get_nodes_dict._cached_nodes = nodes
    return nodes


class MMHeapInfo(gdb.Command):
    """Show basic heap information"""

    def __init__(self):
        super().__init__("mm heap", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        for heap in get_heaps():
            regions = [(start.address, end.address) for start, end in heap.regions]
            gdb.write(f"{heap} - has {len(list(heap.nodes))} nodes, regions:")
            gdb.write(" ".join(f"{hex(start)}~{hex(end)}" for start, end in regions))
            gdb.write("\n")


@autocompeletion.complete
class MMPoolInfo(gdb.Command):
    """Show basic heap information"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Dump memory pool information.")
        parser.add_argument(
            "--heap",
            type=str,
            metavar="file",
            help="Which heap's pool to show",
            default=None,
        )
        return parser

    def __init__(self):
        super().__init__("mm pool", gdb.COMMAND_USER)
        utils.alias("mempool", "mm pool")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        try:
            args = self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

        heaps = [utils.parse_and_eval(args.heap)] if args.heap else get_heaps()
        if not (pools := list(get_pools(heaps))):
            gdb.write("No pools found.\n")
            return

        count = len(pools)
        gdb.write(f"Total {count} pools\n")

        name_max = max(len(pool.name) for pool in pools) + 11  # 11: "@0x12345678"
        formatter = (
            "{:>%d} {:>11} {:>9} {:>9} {:>9} {:>9} {:>9} {:>9} {:>9} {:>9}\n" % name_max
        )
        head = (
            "",
            "total",
            "blocksize",
            "bsize",
            "overhead",
            "nused",
            "nfree",
            "nifree",
            "nwaiter",
            "nqueue",
        )

        gdb.write(formatter.format(*head))
        for pool in pools:
            gdb.write(
                formatter.format(
                    f"{pool.name}@{pool.address:#x}",
                    pool.total,
                    pool.blocksize,
                    pool.size,
                    pool.overhead,
                    pool.nused,
                    pool.nfree,
                    pool.nifree,
                    pool.nwaiter,
                    pool.nqueue,
                )
            )



############################################################################
# tools/pynuttx/nxgdb/memclassifier/default.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from nxgdb.memclassifier.utils import (
    backtrace_function_name_contain,
    backtrace_function_name_equal,
    backtrace_function_name_startswith,
)


def is_quickjs(mb):
    for funcname, filename in mb.backtrace():
        if funcname in ["js_def_malloc", "js_def_realloc"]:
            return True
    return False


def is_graphic(mb):
    for funcname, filename in mb.backtrace():
        if funcname in [
            "ft_alloc",
            "ft_realloc",
            "FT_Stream_New",
            "lv_mem_realloc",
            "lv_mem_alloc",
            "lv_malloc",
            "lv_realloc",
            "lv_malloc_zeroed",
            "gui_context_init(bool)",
            "gui_create_widget(void*, char const*)",
            "ferry::WidgetDiv::setStyle(int, char const*)",
            "ferry::WidgetIMP::onInit()",
            "gui_get_taskinfo()",
            "ferry::DomElement::refreshingWidgetStyle()",
            "gui_create_context(bool)",
            "ferry::DomEntity::updateWidgetChildren(ferry::DomEntity**)",
            "ferry::DomElement::setAttr(ferry::ItemKeyValue const*)",
        ]:
            return True

    return False


def is_freetype_cache(mb):
    for funcname, filename in mb.backtrace():
        if funcname in [
            "freetype_image_create_cb",
        ]:
            return True
    return False


def is_image_cache(mb):
    for funcname, filename in mb.backtrace():
        if funcname in [
            "freetype_image_create_cb",
        ]:
            return True
    return False


def is_style(mb):
    for funcname, filename in mb.backtrace():
        if any(funcname.startswith(func) for func in ["ferry::style::"]):
            return True
    return False


categories = {
    "quickjs": is_quickjs,
    "graphic": {
        "judger": is_graphic,
        "subcategories": {
            "lv_malloc": backtrace_function_name_equal(
                ["lv_malloc_core", "lv_realloc_core"]
            ),
            "ft_realloc ": backtrace_function_name_equal(["ft_realloc "]),
            "freetype_cache": {"judger": is_freetype_cache},
            "freetype": backtrace_function_name_equal(["ft_alloc"]),
            "font": backtrace_function_name_equal(["ferry::Font::"]),
            "textBase": backtrace_function_name_startswith(["ferry::TextBase::"]),
            "image_cache": {"judger": is_image_cache},
            "widget": backtrace_function_name_startswith(["ferry::Widget"]),
            "string_with_Kid": backtrace_function_name_startswith(
                ["ferry::stringToKidInitialize", "ferry::kidToStringInitialize"]
            ),
            "ImageCacheManager": backtrace_function_name_startswith(
                ["ferry::ImageCacheManager::", "gui_init_image_cachemanager"]
            ),
            "ImageProvider": backtrace_function_name_startswith(
                ["ferry::ImageProvider::"]
            ),
            "Yoga": backtrace_function_name_startswith(["YGNode", "facebook::yoga::"]),
            "onPreset": backtrace_function_name_startswith(
                ["ferry::WidgetIMP::onPreSet"]
            ),
        },
    },
    "style": {
        "judger": backtrace_function_name_startswith(
            ["ferry::style::", "ferry::DomElement::computeStyle"]
        ),
        "subcategories": {
            "StrRefHeader": backtrace_function_name_startswith(
                ["ferry::style::StrRefHeader::From"]
            ),
            "computeStyle": backtrace_function_name_contain(["computeStyle"]),
            "StringPool": backtrace_function_name_startswith(
                ["ferry::style::StringPool::StringPool"]
            ),
            "selector": backtrace_function_name_startswith(
                ["ferry::style::StyleManagerBuilder::addSelector"]
            ),
            "StyleManager": backtrace_function_name_startswith(
                ["ferry::style::StyleManager::StyleManager"]
            ),
        },
    },
    "libuv": backtrace_function_name_equal(["uv__realloc", "uv_pipe", "uv_loop_init"]),
    "dom": {
        "judger": backtrace_function_name_startswith(["js_dom_create", "ferry::PbDom"]),
        "subcategories": {
            "create_element_from_js": backtrace_function_name_startswith(
                ["create_element_from_js", "js_dom_create_element"]
            ),
        },
    },
    "curl": backtrace_function_name_startswith(["Curl_"]),
    "native_proxy": backtrace_function_name_startswith(
        [
            "AIOTJS::js_nativeproxy_get",
            "AIOTJS::ComponentCallHook::pauseTracking",
        ]
    ),
    "rapidjson": backtrace_function_name_startswith(
        ["rapidjson::ParseResult", "rapidjson::GenericReader"]
    ),
    "feature_require": backtrace_function_name_contain(
        ["ferry::__app_require", "_onRequired"]
    ),
    "bundle": backtrace_function_name_startswith(["ferry::ZipFile::ZipFile"]),
    "binder": backtrace_function_name_equal(
        ["binder_mmap", "android::RefBase::RefBase()"]
    ),
    "stack": {
        "judger": backtrace_function_name_startswith(["up_create_stack"]),
        "ignore": True,
    },
    "framework": backtrace_function_name_startswith(
        [
            "ferry::Framework::",
            "ferry::Application::",
            "ferry::__runApplication",
            "ferry::PageLoader::createPage(ferry::PageInfo const&)",
            "ferry::PageNavigator::",
            "ferry::Page::init",
            "os::app::ApplicationThread::ApplicationThread(os::app::Application*)",
            "ferry::AppLoader::initAppEnv(ferry::Application*",
        ]
    ),
    "other": lambda mb: True,
}



############################################################################
# tools/pynuttx/nxgdb/rpmsgtrace2pcap.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
from os import path
from typing import Optional

import gdb
from rpmsgtrace2pcap import note_binary_parser

from . import autocompeletion, utils
from .noteram import NoteRam


@autocompeletion.complete
class NoteToPcapCommand(gdb.Command):
    """GDB command for converting noteram data to pcap format"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-d",
            "--driver",
            type=str,
            default="g_rpmsg_note_driver",
            help="Noteram driver name",
        )
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            metavar="file",
            default="rpmsgtrace2pcap.perfetto",
            help="Output pcap file path",
        )
        parser.add_argument(
            "-s", "--save-raw", type=str, help="Save raw trace data to path"
        )
        parser.add_argument(
            "-f", "--freq", type=float, help="Frequency value for trace processing"
        )
        return parser

    def __init__(self):
        if not utils.get_field_nitems("struct noteram_driver_s", "buffer"):
            return
        super().__init__("rpmsgtrace2pcap", gdb.COMMAND_USER)
        self.noteram = None
        self.parser = self.get_argparser()

    def parse_arguments(self, args):
        try:
            return self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return None

    def convert_notes(
        self,
        driver_name: str,
        out_path: str,
        save_path: Optional[str] = None,
        freq: Optional[float] = None,
    ) -> Optional[bytes]:

        # Initialize converter
        self.noteram = NoteRam(driver_name)
        self.noteram.collect_notes(out_path, save_path)

        # Call note_binary_parser function and pass freq parameter
        note_binary_parser(self.noteram.elf_parser, self.noteram.note_parser, freq=freq)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(args)
        if not args:
            return

        try:
            self.convert_notes(
                args.driver,
                path.abspath(args.output),
                args.save_raw,
                args.freq,
            )
        except Exception as e:
            print(f"Error in note conversion: {e}")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "NoteToPcap Report",
            "summary": "Note to pcap conversion",
            "command": "rpmsgtrace2pcap",
            "result": "info",
            "category": utils.DiagnoseCategory.system,
            "message": gdb.execute("rpmsgtrace2pcap", to_string=True),
        }



#!/usr/bin/python3
############################################################################
# tools/pynuttx/gdbrpc/tests/stresstest.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import os
import queue
import random
import sys
import threading
import time
import unittest
from typing import List

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

import gdbrpc  # noqa: E402

gdb_list = ["bt", "gcore", "info registers", "info threads"]

nxgdb_list = [
    "dump ram",
    "info nxthread",
    "memfind 0x12341234",
    "nxgcore  -p arm-none-eabi-",
    "ps",
    "source tools/pynuttx/tests/runner.py",
    "stack-usage",
]

command_list = nxgdb_list + gdb_list


class ReturnStrValue(gdbrpc.Request):
    def __init__(self):
        super().__init__()

    def __call__(self, event: queue.Queue):
        import nxgdb

        heaps = nxgdb.mm.get_heaps()
        event.put(str(heaps))


class BaseCase(unittest.TestCase):
    """Base"""

    def verify_registers(self, start_registers: str, end_registers: str):
        self.assertNotIn("thread is running", start_registers.lower())
        self.assertNotIn("thread is running", end_registers.lower())
        for start, end in zip(start_registers.splitlines(), end_registers.splitlines()):
            self.assertEqual(start, end)

    def random_continue_and_interrupt(self, client: gdbrpc.Client):
        self.assertEqual(client.call(gdbrpc.ShellExec("interrupt")), "")

        self.assertEqual(client.call(gdbrpc.ShellExec("nxcontinue")), "")
        sleep_time = random.uniform(3, 6)
        time.sleep(sleep_time)
        self.assertEqual(client.call(gdbrpc.ShellExec("interrupt")), "")


class TestClient(BaseCase):
    """Test cases for gdbrpc Client class."""

    def test_client_initialization(self):
        """Test client initialization with default and custom parameters."""
        client = gdbrpc.Client()
        self.assertEqual(client._host, "localhost")
        self.assertEqual(client._port, 20819)
        self.assertFalse(client._connected)

    def test_client_call(self):
        client = gdbrpc.Client()
        self.assertTrue(client.connect())
        self.assertTrue(isinstance(client.call(ReturnStrValue()), str))
        client.disconnect()

    def test_send_continue(self):
        client = gdbrpc.Client()
        self.assertTrue(client.connect())

        self.assertEqual(client.call(gdbrpc.ShellExec("interrupt")), "")

        self.assertEqual(client.call(gdbrpc.ShellExec("nxcontinue")), "")
        sleep_time = random.uniform(3, 6)
        time.sleep(sleep_time)
        self.assertEqual(client.call(gdbrpc.ShellExec("interrupt")), "")

        client.disconnect()

    def test_send_interrupt(self):
        client = gdbrpc.Client()
        self.assertTrue(client.connect())

        for _ in range(15):
            self.random_continue_and_interrupt(client)

        client.disconnect()


class TestSendNxGDB(BaseCase):
    def run_command(self, client: gdbrpc.Client, command: str):
        start_registers: str = client.call(gdbrpc.ShellExec("info registers"))
        client.call(gdbrpc.ShellExec(command))
        end_registers: str = client.call(gdbrpc.ShellExec("info registers"))
        self.verify_registers(start_registers, end_registers)

    def send_commands(self, client: gdbrpc.Client):
        for _ in range(5):
            self.run_command(client, random.choice(command_list))

    def test_send_command(self):
        client = gdbrpc.Client(logLevel=logging.DEBUG)
        self.assertTrue(client.connect())

        self.random_continue_and_interrupt(client)

        for command in command_list:
            self.run_command(client, command)

        client.disconnect()

        for _ in range(5):

            client_total = 60
            clients: List[gdbrpc.Client] = [
                gdbrpc.Client() for _ in range(client_total)
            ]
            for client in clients:
                self.assertTrue(client.connect())

            threads: List[threading.Thread] = []
            for client in clients:
                thread = threading.Thread(target=self.send_commands, args=(client,))
                threads.append(thread)
                thread.start()

            for thread in threads:
                thread.join()

            self.random_continue_and_interrupt(client)

            for client in clients:
                client.disconnect()


if __name__ == "__main__":
    unittest.main()



############################################################################
# tools/pynuttx/nxstub/gdbstub.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import socket
import traceback
from binascii import hexlify, unhexlify
from typing import Union

from . import utils
from .target import Target


class GDBStub:
    def __init__(
        self,
        target: Target,
        port=1234,
        proxymode=False,
    ):
        self.proxymode = proxymode
        self.threads = target.update_threads()
        self.registers = target.switch_thread()
        self.target = target
        self.exiting = False
        self.socket = None
        self.port = port
        self.logger = logging.getLogger(__name__)

    def run(self):
        self.socket = self.wait_connect()
        if self.socket is None:
            return

        while not self.exiting:
            try:
                packet = utils.get_packet(self.socket)
                if packet in b"+-":
                    continue

                self.socket.send(b"+")
                if packet is None:
                    self.logger.info("Connection closed")
                    return

                if self.proxymode:
                    self.forward_packet(packet)
                else:
                    self.process_packet(packet)

            except Exception as e:
                self.logger.error(f"Error in stub thread: {e} {traceback.format_exc()}")
        self.logger.info("Stub thread exited")

    def wait_connect(self) -> socket.socket:
        listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        port = self.port
        try:
            listener.bind(("localhost", port))
            self.logger.info(f"Listening on localhost:{port}")
            listener.listen(1)
        except socket.error as e:
            self.logger.error(
                f"Cannot listen on localhost:{port}: error {e[0]} - {e[1]}"
            )
            return None

        client, addr = listener.accept()
        self.logger.info(f"Client connected from {addr[0]}:{addr[1]}")
        listener.close()
        return client

    def send_raw_packet(self, data: bytes, nowait=False) -> bool:
        checksum = sum(data) & 0xFF
        self.socket.send(b"$")
        self.socket.send(data)
        self.socket.send(b"#")
        self.socket.send(b"%02x" % checksum)
        self.logger.debug(f"Sent packet: {data}")
        if nowait:
            return True
        ack = self.socket.recv(1)
        return ack == b"+"

    def send_packet(self, packet: Union[bytes, str], nowait=False) -> None:
        if isinstance(packet, str):
            packet = packet.encode("ascii")

        output = utils.encode_packet(packet)
        return self.send_raw_packet(output, nowait)

    def send_unsupported(self):
        self.send_packet("")

    def process_packet(self, packet: bytes):
        self.logger.debug(f"Process packet: {packet}")
        attribute = "handle_" + chr(packet[0])
        handler = {
            "handle_?": self.handle_questionmark,
            "handle_\x03": self.handle_etx,
        }.get(attribute)

        if not handler and not hasattr(self, attribute):
            self.logger.error(f"Unsupported packet: {packet}")
            self.send_unsupported()
            return

        handler = handler or getattr(self, attribute)

        try:
            handler(packet)
        except Exception as e:
            self.logger.error(f"Error packet{packet}: {e}\n {traceback.format_exc()}")
            self.send_packet("EF1")

    def forward_packet(self, packet: bytes):
        packet_type = packet[0]
        if packet_type in b"iIzZsSmMxX":
            try:
                self.logger.debug(f"Forwarding packet: {packet[:20]}...")
                response = self.target.send_packet(packet)
                self.logger.debug(f"Received response: {response and response[:20]}...")
            except Exception as e:
                self.logger.error(f"Error forwarding packet: {e}")
                return

            if response:
                # The response is the raw packet received via RSP, excluding start, end and checksum
                self.send_raw_packet(response, nowait=True)
            else:
                self.send_packet("")
        elif packet_type in b"?vqgpPTHc\x03D":
            self.process_packet(packet)
        else:
            self.logger.debug(f"Ignore packet: {packet}")
            self.send_unsupported()

    def handle_q(self, packet: bytes):
        packet = packet.decode("ascii")
        if packet.startswith("qSupported"):
            self.send_packet("PacketSize=FFFFFF;binary-upload+")
        elif packet.startswith("qC"):
            pid = self.target.current_thread()
            self.logger.debug(f"Current thread: {pid}")
            self.send_packet(f"QC{pid:x}")
        elif packet.startswith("qfThreadInfo"):
            reply = "".join(f"{thread.pid:x}," for thread in self.threads)
            reply = "m" + reply[:-1] if reply else "l"
            self.send_packet(reply)
        elif packet.startswith("qsThreadInfo"):
            self.send_packet("l")
        elif packet.startswith("qThreadExtraInfo"):
            pid = int(packet.split(",")[1], 16)
            info = next((t for t in self.threads if t.pid == pid), None)
            info = f"{info.name},{info.state}" if info else f"Invalid PID {pid}"
            self.send_packet(hexlify(info.encode("ascii")))
        elif packet.startswith("qRcmd"):
            try:
                _, command = packet.split(",")
                command = unhexlify(command)
                response = self.target.monitor_command(command)
                if response is None:
                    self.send_unsupported()
                    return
            except Exception as e:
                self.logger.error(f"Error executing monitor command: {e}")
                # Note that older GDB may treat it as normal output instead of error, but nothing hurts.
                response = f"E.{e}\n"

            if isinstance(response, str):
                response = response.encode("ascii")
            self.send_packet(hexlify(response))
        else:
            self.send_unsupported()

    def handle_v(self, packet: bytes):
        if packet.startswith(b"vMustReplyEmpty"):
            self.send_packet("")
        else:
            self.send_unsupported()

    def handle_questionmark(self, packet: bytes):
        running = next((t for t in self.threads if t.state == "Running"), None)
        # FIXME t.state may be a number when g_statesnames gets optimized out.
        # That's fine for now since we can identify the running thread manually.
        if running:
            self.registers = self.target.switch_thread(running.pid)
            self.send_packet(f"T02thread:{running.pid:x};")
        else:
            self.send_packet("S05")

    def handle_g(self, packet: bytes):
        self.send_packet(self.registers.to_g())

    def handle_p(self, packet: bytes):
        packet = packet.decode("ascii")
        regnum = int(packet[1:], 16)
        reg = self.registers.get(regnum=regnum)
        self.send_packet(hexlify(bytes(reg)) if reg else b"xx" * 4)

    def handle_P(self, packet: bytes):
        packet = packet.decode("ascii")
        regnum, value = packet[1:].split("=")
        regnum = int(regnum, 16)
        value = unhexlify(value)
        self.registers.set(regnum=regnum, value=value)
        self.send_packet("OK")

    def handle_x(self, packet: bytes):
        packet = packet.decode("ascii")
        addr, length = packet[1:].split(",")
        data = self.target.memory_read(int(addr, 16), int(length, 16))
        self.send_packet(b"b" + bytes(data))

    def handle_X(self, packet: bytes):
        addr_and_length, data = packet[1:].split(b":")
        addr_and_length = addr_and_length.decode("ascii")
        addr, length = addr_and_length.split(",")
        if int(length, 16) != len(data):
            self.send_packet("E01")
            return
        ok = self.target.memory_write(int(addr, 16), data)
        self.send_packet("OK" if ok else "")

    def handle_m(self, packet: bytes):
        packet = packet.decode("ascii")
        addr, length = packet[1:].split(",")
        reply = self.target.memory_read(int(addr, 16), int(length, 16))
        reply = hexlify(reply)
        self.send_packet(reply)

    def handle_M(self, packet: bytes):
        packet = packet.decode("ascii")
        addr, length_and_data = packet[1:].split(",")
        length, data = length_and_data.split(":")
        if int(length, 16) != len(data) // 2:
            self.send_packet("E01")
            return

        ok = self.target.memory_write(int(addr, 16), unhexlify(data))
        self.send_packet("OK" if ok else "")

    def handle_etx(self, packet: bytes):
        self.logger.info("Ctrl-C received")
        if self.proxymode:
            self.target.stop()
        else:
            self.send_packet("S00")

    def handle_k(self, packet: bytes):
        self.exiting = True
        self.logger.info("Kill request received")

    def handle_T(self, packet):
        self.send_packet("OK")

    def handle_H(self, packet):
        pid = int(packet[2:], 16)
        registers = self.target.switch_thread(pid)
        if not registers:
            self.send_packet("E01")
            return
        self.registers = registers
        self.send_packet("OK")

    def handle_c(self, packet):
        if hasattr(self.target, "cont"):

            def stopped(stopreason):
                # FIXME callback from another thread
                self.logger.info(f"Target stopped: {stopreason}")
                self.handle_questionmark(None)

            # Notify the stub when it's stopped
            self.target.cont(stopped)
        else:
            self.send_unsupported()



############################################################################
# tools/pynuttx/gdbrpc/utils.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import queue
import socket
import struct
import subprocess
import threading
from enum import IntEnum
from typing import Any

DEFAULT_TIMEOUT = 300


def recv_all(connection: socket.socket, length: int, logger: logging.Logger) -> bytes:
    data = b""
    while len(data) < length:
        chunk = connection.recv(length - len(data))
        if not chunk:
            logger.error("Socket connection broken during receive")
            raise ConnectionError("Socket connection broken during receive")
        data += chunk
    return data


def socket_recv(connection: socket.socket, logger: logging.Logger) -> bytes:
    length_data = recv_all(connection, 4, logger)
    # https://docs.python.org/3.10/library/struct.html#format-characters
    data_length = struct.unpack("!I", length_data)[0]
    logger.debug(f"Expecting to receive {data_length} bytes")

    return recv_all(connection, data_length, logger)


def send_all(connection: socket.socket, data: bytes, logger: logging.Logger) -> None:
    total_sent = 0
    while total_sent < len(data):
        sent = connection.send(data[total_sent:])
        if sent == 0:
            logger.error("Socket connection broken during send")
            raise RuntimeError("Socket connection broken")
        total_sent += sent


def socket_send(connection: socket.socket, data: bytes, logger: logging.Logger) -> None:
    try:
        data_length = len(data)
        length_prefix = struct.pack("!I", data_length)
        logger.debug(f"Sending {data_length} bytes")

        send_all(connection, length_prefix, logger)
        send_all(connection, data, logger)

    except Exception as e:
        logger.error(f"Error sending response: {e}")
        raise


class PacketStatus(IntEnum):
    HAS_CALLBACK = 0
    NO_CALLBACK = 1
    PYTHON_VERSION_MISMATCH = 2


class Response:
    def __init__(self, tag: int, payload: Any):
        self.tag = tag
        self.payload = payload


class Request:
    def __init__(self):
        self.tag = id(self)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        raise NotImplementedError("Subclasses must implement this method")


class PostRequest(Request):
    def __init__(self):
        super().__init__()
        self.finish = threading.Event()

    def __call__(self, argument: Any):
        raise NotImplementedError("Subclasses must implement this method")


class ShellExec(Request):
    def __init__(self, command: str):
        super().__init__()

        self.is_gdb_command = True
        command = command.strip()

        if command.startswith("!"):
            command = command[1:].strip()
            self.is_gdb_command = False
        elif command.startswith("shell"):
            command = command[5:].strip()
            self.is_gdb_command = False

        self.command = command

    def _run_shell_command(self, command) -> str:
        try:
            process = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
                universal_newlines=True,
            )

            result = ""

            while True:

                stdout_line = process.stdout.readline()
                stderr_line = process.stderr.readline()

                if stdout_line:
                    result += f"{stdout_line}"
                if stderr_line:
                    result += f"{stderr_line}"

                if (
                    stdout_line == ""
                    and stderr_line == ""
                    and process.poll() is not None
                ):
                    break

            return result
        except Exception as e:
            return f"Error executing command '{command}': {e}"

    def __call__(self, queue: queue.Queue):
        import gdb

        try:
            if self.is_gdb_command:
                out = gdb.execute(f"{self.command}", to_string=True)
            else:
                out = self._run_shell_command(self.command.split())
        except Exception as e:
            out = e
        queue.put(out)



############################################################################
# tools/gdb/nuttxgdb/protocols/mm.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from typing import List

from .value import Value


class ProcfsMeminfoEntry(Value):
    """struct procfs_meminfo_entry_s"""

    name: Value
    heap: Value
    next: ProcfsMeminfoEntry


class MMAllocNode(Value):
    """struct mm_allocnode_s"""

    preceding: Value
    size: Value
    pid: Value
    seqno: Value
    backtrace: Value


class MMFreeNode(Value):
    """struct mm_freenode_s"""

    preceding: Value
    size: Value
    pid: Value
    seqno: Value
    backtrace: Value
    flink: MMFreeNode
    blink: MMFreeNode


class MMHeap(Value):
    """struct mm_heap_s"""

    mm_lock: Value
    mm_heapsize: Value
    mm_maxused: Value
    mm_curused: Value
    mm_heapstart: List[MMAllocNode]
    mm_heapend: List[MMAllocNode]
    mm_nregions: Value
    mm_nodelist: Value


class MemPool(Value):
    """struct mempool_s"""

    initialsize: Value
    interruptsize: Value
    expandsize: Value
    wait: Value
    priv: Value
    alloc: Value
    free: Value
    check: Value
    ibase: Value
    queue: Value
    iqueue: Value
    equeue: Value
    nalloc: Value
    lock: Value
    waitsem: Value
    procfs: Value


class MemPoolMultiple(Value):
    """struct mempool_multiple_s"""

    pools: List[MemPool]
    npools: Value
    expandsize: Value
    minpoolsize: Value
    arg: Value
    alloc: Value
    alloc_size: Value
    free: Value
    alloced: Value
    delta: Value
    lock: Value
    chunk_queue: Value
    chunk_size: Value
    dict_used: Value
    dict_col_num_log2: Value
    dict_row_num: Value
    dict: Value


class MemPoolBlock(Value):
    """struct mempool_record_s"""

    magic: Value
    pid: Value
    seqno: Value
    backtrace: Value
